/*****************************************************************************
*  (c) Copyright 1994 Wide Area Information Servers, Inc                     *
*      All rights reserved.                                                  *
*                                                                            *
*  This code can be freely copied and used for any purpose as long as        *
*  this notice and copyright notice is included without modification         *
*  at the top of the file.                                                   *
*                                                                            *
*****************************************************************************/

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Module:   demo-main.c

 Purpose:  The client-main module provides an example program of how to use
           the client protocol toolkit library.  It makes use of the
           Application Programmers Interface (API) described by the following
           header files:
                         nConnMgr.h
                         appConnMgr.h
                         appConnCli.h
---------------------------------------------------------------------------*/

#include <appConnCli.h>

/* For demonstration purposes, the buffer size is small.  In practice, 
   a buffer size of 100000L or more will improve performance.  */
#define BUFFER_SIZE 10000L

#define TEST_SERVER "wais.com"
#define TEST_CLIENT "Demo Client"
#define TEST_PORT 2040L

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   init
 Purpose:    Initializes the application connection, given by the appConn,
             conn.
 Called by:  main
 Parameters: conn   -- input; a handle to an appConn. It holds
                       information about the state of the connection;
                       may be modified by this function.
             client -- input; a null-terminated string whose value is
                       the client performing the request.
             diags  -- output; a handle to a list of diagnostics,
                       where each item in the list is of type 
                       appConnDiag (see appConnMgr.c).  If the
                       operation was unsuccessful, a list of diags
                       are returned in this parameter.
 Returns:    boolean, true(1) if successful, false(0) otherwise.
---------------------------------------------------------------------------*/
boolean
init(appConn* conn,
     char*    client,
     list*    diags)
{
  long pduType = appConnUnknown;
  appConnStatus as;

  appConnInitReqWrite(conn,client,"","",NULL,diags);

  if (*diags != NULL) return(false);

  while (((as = appConnPduInfo(conn,&pduType,NULL,diags)) 
	  == appConnStatusNotFinished) && (*diags == NULL)) ;

  if (*diags != NULL) return(false);
  
  if (pduType != appConnInitResp) return(false);

  while (((as = appConnInitRespRead(conn,NULL,diags)) 
	  == appConnStatusNotFinished) && (*diags == NULL)) ;

  if (*diags != NULL) return(false);

  return(true);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   search
 Purpose:    Performs a search of a query on a database using the application
             connection, conn.
 Called by:  Demo functions: testSearchAndRetrieveByRecId
                             testSearchWithRFAndRetrieveByPosition
 Parameters: conn         -- input; a handle to an appConn. It holds
                             information about the state of the connection;
                             may be modified by this function.
             maxHits      -- input; a long integer whose value is the
                             maximum number of results the server is 
                             requested to generate. maxHits results are
                             to be stored on the server for the duration
                             of this search.
             dbName       -- input; a null-terminated string containing
                             the name of the database to search.
             qry          -- input; a handle to an appConnQry data type
                             containing the user's query and is created
                             using the appQryAddTerm() function.
             qryTermsUsed -- output; a handle to a list of null-terminated
                             character strings.  Each character string
                             contains the query terms used by the server
                             in the search; these terms may or may not
                             have been specified in the original query.
                             The terms are dependent on the behavior of the
                             server and the contents of the database.  
                             Stop words may have been eliminated, right-
                             truncated words may have been expanded, and
                             synonyms may have been included.
             hits         -- output; a handle to a list.  Each item
                             in the list is of type appConnHeadline, and
                             represents one element of the result set.  
                             The number of hits returned may be less than
                             or equal to the number of hits requested.
             diags        -- output; a handle to a list of diagnostics,
                             where each item in the list is of type 
                             appConnDiag (see appConnMgr.c).  If the
                             operation was unsuccessful, a list of diags
                             are returned in this parameter.
 Returns:    boolean, true(1) if successful, false(0) otherwise.
---------------------------------------------------------------------------*/
boolean
search(appConn* conn,
       long maxHits,
       char* dbName,
       appConnQry* qry,
       list* qryTermsUsed,
       list* hits,
       list* diags)
{
  long pduType = appConnUnknown;
  appConnStatus as;
  long totalNumHits, currentNumHits, requestedNumHits;

  /* Send the search request to the remote system. */
  appConnSearchReqWrite(conn,maxHits,maxHits,rankByServerDefault,
			dbName,qry,NULL,diags);

  if (*diags != NULL) return(false);

  /* Receive the response. */
  while (((as = appConnPduInfo(conn,&pduType,NULL,diags)) 
	  == appConnStatusNotFinished) && (*diags == NULL)) ;

  if (*diags != NULL) return(false);
     
  /* Make sure the response has the correct packet type. */
  if (pduType != appConnSearchResp) return(false);

  /* Read the response and get the headliness. */
  while (((as = appConnSearchRespRead(conn,&totalNumHits,qryTermsUsed,
				      hits,NULL,diags))
	  == appConnStatusNotFinished) && (*diags == NULL)) ;

  if (*diags != NULL) return(false);

  /* Remember how many headlines we currently have,
     and how many we want to receive. */
  currentNumHits = length(*hits);
  maxHits = (maxHits < 0) ? MAXLONG : maxHits;
  requestedNumHits = MIN(maxHits,totalNumHits);

  /* If not enough headlines were returned, keep asking for more headlines. */
  while (currentNumHits < requestedNumHits)
    
   { 
     /* Send a request for the next set of headlines. */
     as = appConnNextHeadlinesReqWrite(conn,
				       1+currentNumHits,
				       requestedNumHits - currentNumHits,
				       NULL,NULL);

     if (as == appConnStatusUndefinedFunction) break;
     if (*diags != NULL) return(false);

     /* Receive the response. */
     while (((as = appConnPduInfo(conn,&pduType,NULL,diags)) 
	     == appConnStatusNotFinished) && (*diags == NULL)) ;
     
     if (*diags != NULL) return(false);

     /* Make sure the response has the correct packet type. */
     if (pduType != appConnNextHeadlinesResp) return(false);

     /* Read the response and get the headliness. */
     while (((as = appConnNextHeadlinesRespRead(conn,hits,NULL,diags)) 
	     == appConnStatusNotFinished) && (*diags == NULL)) ;

     if (*diags != NULL) return(false);

     /* Update the current headline count. */
     currentNumHits = length(*hits);
   }

  return(true);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   retrieve
 Purpose:    Performs a retrieval of a document.  The document can be retrieved
             by specifying the position in a result set from a previous search,
             by a local record identifier, or by a URx (Uniform Resource
             Identifier).
 Called by:  Test functions: testSearchAndRetrieveByRecId
                             testSearchWithRFAndRetrieveByPosition
 Parameters:  conn          -- input; a handle to an appConn. It holds
                               information about the state of the connection;
                               may be modified by this function.
              retrievalType -- input;; an long integer used to specify the 
                               retrieval method.  Must be one of the following
                               defined in the file "appConnMgr.h":
                                   appConnRetrieveByURx
                                   appConnRetrieveByRecId
                                   appConnRetrieveByPosition
              resSetPos     -- input; a long integer used to specify the 
                               position in the result set of the document
                               that is requested.  Applicable only for the
                               retrieval type of appConnRetrieveByPosition.
              recId         -- input; a handle to an 'any' datatype;
                               specifies the record identifier of the
                               requested document.
              docId         -- input; a null-terminated character string;
                               specifies the record identifier or URx
                               of the requested document.
              elemId        -- input; a null-terminated character string;
                               specifies the element identifier of the
                               requested document.
              varId         -- input; a null-terminated character string; 
                               specifies the variant identifier of the
                               requested document.
              dbName        -- input; a null-terminated string.  It
                               contains the name of the database from which
                               to retrieve the data.  Only required when
                               the retrieval type is appConnRetrieveByUrx
                               or appConnRetrieveByRecId.
              data          -- output; a pointer to a null-terminated character
                               string containing the requested data.
              dataLength    -- output; a pointer to a long integer whose value
                               contains the number of bytes in data.
              diags         -- output; a handle to a list of diagnostics,
                               where each item in the list is of type 
                               appConnDiag (see appConnMgr.c).  If the
                               operation was unsuccessful, a list of diags
                               are returned in this parameter.
 Returns:    boolean, true(1) if successful, false(0) otherwise.
---------------------------------------------------------------------------*/
boolean
retrieve(appConn* conn,
	 long     retrievalType,
	 long     resSetPos,
	 any*     recId,
	 char*    docId,
	 char*    elemId,
	 char*    varId,
	 char*    dbName,
	 char**   data,
	 long*    dataLength,
	 list*    diags)
{
  long pduType = appConnUnknown;
  long i, offset, len;
  char* dataReturned = NULL;
  appConnDiag* diag;
  appConnStatus as;

  *dataLength = 0;
  for(offset=0L,len=conn->negotiatedBufSz-500L;len>0L;offset+=len)
   {
     as = appConnRetrieveReqWrite(conn,retrievalType,resSetPos,
				  recId,docId,elemId,varId,dbName,
				  offset,len,NULL,diags);

     if (as != appConnStatusOK) { appConnStatusLogStatus(as); return(false); }
     if (*diags != NULL) return(false);

     while (((as = appConnPduInfo(conn,&pduType,NULL,diags)) 
	     == appConnStatusNotFinished) && (*diags == NULL)) ;
     
     if (pduType != appConnRetrieveResp) return(false);
     if (*diags != NULL) return(false);

     while (((as = appConnRetrieveRespRead(conn,&dataReturned,&len,NULL,diags))
	     == appConnStatusNotFinished) && (*diags == NULL)) ;

     *dataLength += len;
     *data = safeRealloc(*data,(size_t)((*dataLength)+1));
     memcpy((*data)+offset,dataReturned,(size_t)len);
     safeFree(dataReturned);

     for(i=0,diag = nth(*diags,0);i<length(*diags);i++,diag = nth(*diags,i))
       { if (diag->code == appConnDiagRequestOutOfRange) 
	   return(true);
	 return(false);
       }
   }

  return(true);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   displaySearchResults
 Purpose:    Displays search results (hits) to stdout.
 Called by:  Test functions: testSearchAndRetrieveByRecId
                             testSearchWithRFAndRetrieveByPosition
 Parameters: hits -- input; a list of hits where each hit is of datatype
                     appConnHeadline.
 Returns:    void
---------------------------------------------------------------------------*/
void
displaySearchResults(list hits)
{
  appConnHeadline* hit;
  appConnElement*  element;
  appConnVariant*  variant;
  long i, j, k;

  for (i=0; i<length(hits); i++)
   { 
     long count = 0;
     hit = nth(hits,i);

     /* Print elements and variants. */
     printf("%3ld   score %4ld %s\n",hit->rank,hit->score,hit->headline);
	
     /* Print each component */
     for(j=0; j<length(hit->elements); j++)
      { 
	element = nth(hit->elements,j);

	/* If only one variant exists. */
	if (length(element->variants) == 1)
	 { 
	   variant = first(element->variants);
	   printf("%3ld.%ld %25s len %7ld\n",hit->rank,count++,
		  element->name,variant->size);
	 }
	else 
	  /* Print all the variants of the component. */
	 { 
	   long k;
	   printf("%25s\n",element->name);
	   for(k=0; k<length(element->variants); k++)
	    { 
	      variant = nth(element->variants,k);
	      printf("%3ld.%ld len %7ld\n",hit->rank,count++,
		     variant->name,variant->size);
	    }
	 }
      }
   }
}




/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   displayRetrievalResult
 Purpose:    Display the retrieval result to stdout.
 Called by:  Test functions: testSearchAndRetrieveByRecId
                             testSearchWithRFAndRetrieveByPosition
 Parameters: data -- input; a character string containing the retrieved data
                     to be displayed.
             size -- input; the length of the data character string.
 Returns:    void
---------------------------------------------------------------------------*/
void displayRetrievalResult(char* data,
			    long  size)
{
  fwrite(data,1,size,stdout);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   displayDiagsAndCleanup
 Purpose:    Display the diagnostics to stdout, and frees qry, hits, data, 
             and diags.
 Called by:  Test functions: testSearchAndRetrieveByRecId
                             testSearchWithRFAndRetrieveByPosition
 Parameters: qry   -- input; a handle to an appConnQry data type created
                      by the appQryAddTerm() function.
             hits  -- input; a handle to a list.  Each item in the list is
                      of type appConnHeadline.
             data  -- input; a character string containing the retrieved data
                      to be freed.
             diags -- output; a handle to a list of diagnostics, where each
                      item in the list is of type appConnDiag (see 
                      appConnMgr.c).  If the operation was unsuccessful, 
                      a list of diags are returned in this parameter.
 Returns:    void
---------------------------------------------------------------------------*/
void displayDiagsAndCleanup(appConnQry* qry,
			    list        hits,
			    char*       data,
			    list        diags)
{
  appConnDiag* diag;
  long i;

  for(i=0L,diag = nth(diags,0L);i<length(diags);i++,diag = nth(diags,i))
    printf("\nDiagnostic code %ld, %s\n",diag->code,diag->info);

  /* Clean up. */
  appConnQryFree(qry);

  /* Free the hits from the search. */
  for(i=0L; i<length(hits); i++)
    appConnHeadlineFree(nth(hits,i));
  freeList(hits);

  safeFree(data);

  /* Free the diags. */
  for(i=0L; i<length(diags); i++)
    appConnDiagFree(nth(diags,i));
  freeList(diags);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   testSearchAndRetrieveByRecId
 Purpose:    Demonstrates a simple search and retreival, and displays the 
             results.  Also displays the diagnostics.
 Called by:  main
 Parameters: conn -- input; a handle to an appConn. It holds information 
                     about the state of the connection; may be modified by
                     this function.
 Returns:    void
---------------------------------------------------------------------------*/
void
testSearchAndRetrieveByRecId(appConn* conn)
{
  long i;

  char* dbName = "smithsonian-pictures";
  char* qryTxt = "gorilla OR mal*";

  appConnQry* qry = NULL;
  appConnHeadline* hit;
  appConnElement* element;
  appConnVariant* variant;

  list hits = NULL;
  list diags = NULL;
  char* data = NULL;
  long size = 0L;

  long hitNo = 0L;
  long elementNo = 0L;
  long variantNo = 0L;

  /* Create query, search, and display results. */
  printf("\n\nTest search on database %s with query (%s).\n",dbName,qryTxt);
  qry = appConnQryMakeLeafFreeText(qryTxt);
  if (search(conn,100L,dbName,qry,NULL,&hits,&diags))
    displaySearchResults(hits);
  else 
   { 
     printf("\nSearch failed.\n");
     displayDiagsAndCleanup(qry,hits,NULL,diags);
     return;
   }

  /* Select the hit, element, and variant for retrieval. */
  hit = nth(hits,hitNo);
  
  if (hit == NULL)
   { /* didn't get a response back from the server */
     printf("\nSearch returned no hits.\n");
     displayDiagsAndCleanup(qry,hits,NULL,diags);
     return;
   }

  element = nth(hit->elements,elementNo);
  variant = nth(element->variants,variantNo);

  /* Retrieve by record identifier and display result. */
  printf("\n\nTest retrieval of headline %ld, elem %ld, var %d by recId.\n",
	 hitNo,elementNo,variantNo);
  if (retrieve(conn,appConnRetrieveByRecId,0L,hit->recId,NULL,
	       element->name,variant->name,dbName,&data,&size,&diags))
    displayRetrievalResult(data,size);
  else
    printf("\nRetrieval failed.\n");

  displayDiagsAndCleanup(qry,hits,data,diags);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   testSearchWithRFAndRetrieveByPosition
 Purpose:    Demonstrates a search using relevance feedback containing text.
             The text is a portion of a document that the user feeds back
             in a search to obtain more documents similar to the words in
             the text.  Next, demonstrates a search using relevance feedback
             specified with a local record identifier.  And finally, a
             retrieval is performed by specifying the desired document by
             it's position in the result set.
 Called by:  main
 Parameters: conn -- input; a handle to an appConn. It holds information 
                     about the state of the connection; may be modified by
                     this function.
 Returns:    void
---------------------------------------------------------------------------*/
void
testSearchWithRFAndRetrieveByPosition(appConn* conn)
{
  long i;

  char* dbName = "smithsonian-pictures";
  char* freeText = "gorilla OR mal*";
  char* rfDocText = 
"     Astronomers are gaining in their search for planets orbiting\n\
     stars other than our own sun.  The huge Arecibo Observatory,\n\
     built into a mountainside in Puerto Rico, was used to detect\n\
     the most recent evidence of planets outside the solar system.\n\
";

  appConnQry* qry1 = NULL;
  appConnQry* qry2 = NULL;
  appConnHeadline* hit;
  appConnElement* element;
  appConnVariant* variant;

  appConnHeadline* rfHit;
  appConnElement* rfElement;
  appConnVariant* rfVariant;

  list hits1 = NULL;
  list hits2 = NULL;
  list diags = NULL;
  char* data = NULL;
  long size = 0L;

  long hitNo = 4L;
  long elementNo = 0L;
  long variantNo = 0L;
  long rfHitNo = 3L;

  /* Search. */
  printf("\n\nTest search on database %s with free text (%s)",dbName,freeText);
  printf("\nOR'd with relevance feedback text:\n%s",rfDocText);
  qry1 = appConnQryMakeNode(qryNodeOr,
			    appConnQryMakeLeafFreeText(freeText),
			    appConnQryMakeLeafRFDocText(rfDocText));
  if (search(conn,100L,dbName,qry1,NULL,&hits1,&diags))
    displaySearchResults(hits1);
  else 
   {
     printf("\nSearch failed.\n");
     displayDiagsAndCleanup(qry1,hits1,NULL,diags);
     return;
   }

  /* Search with relevance feedback by RecId. */
  printf("\n\nTest search on database %s with free text (%s)",dbName,freeText);
  printf("\nOR'd with relevance feedback by RecId from hit %d.\n",rfHitNo);
  rfHit = nth(hits1,rfHitNo);

  if (rfHit == NULL)
   { /* nothing to feed back */
     printf("\nCan't do RF, since previous search returned no hits.\n");
     displayDiagsAndCleanup(qry1,hits1,NULL,diags);
     return;
   }

  rfElement = nth(rfHit->elements,0L);
  rfVariant = nth(rfElement->variants,0L);
  qry2 = appConnQryMakeNode(qryNodeOr,
			    appConnQryMakeLeafFreeText(freeText),
			    appConnQryMakeLeafRFRecId(rfHit->recId,
						      rfElement->name,
						      rfVariant->name,
						      0L,50L));
  if (search(conn,100L,dbName,qry2,NULL,&hits2,&diags))
    displaySearchResults(hits2);
  else 
   {
     printf("\nSearch failed.\n");
     displayDiagsAndCleanup(qry1,hits1,NULL,NULL);
     displayDiagsAndCleanup(qry2,hits2,NULL,diags);
     return;
   }

  /* Select the hit, element, and variant for retrieval. */
  hit = nth(hits2,hitNo);
  element = nth(hit->elements,elementNo);
  variant = nth(element->variants,variantNo);

  /* Retrieve by position in the result set. */
  printf("\nTest retrieval of headline %ld, elem %ld, var %d by position.\n",
	 hitNo,elementNo,variantNo);
  if (retrieve(conn,appConnRetrieveByPosition,hitNo,hit->recId,NULL,
	       element->name,variant->name,NULL,&data,&size,&diags))
    displayRetrievalResult(data,size);
  else
    printf("\nRetrieval failed. (NOTE - the 1988 version of does not support retrieval\n by position, so failure is expected there.)\n");

  displayDiagsAndCleanup(qry1,hits1,NULL,NULL);
  displayDiagsAndCleanup(qry2,hits2,data,diags);
}


/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   main
 Purpose:    Demonstrates how to use the client API functions.  Opens a 
             single network connection of type "TCP", and an application
             connection of "Z3950V2".  Next performs initialization, runs
             a series of demonstration search and retreival tests, and then
             performs cleanup.
 Parameters: none
 Returns:    void
---------------------------------------------------------------------------*/
#define NUMBER_OF_DEMO_PROTOCOLS 2
void
main()
{
  char* host = TEST_SERVER;
  long port = TEST_PORT;
  char buf[BUFFER_SIZE];
  long i, j;
  char* nConnTypeName = "TCP";
  char* appConnTypeNameList[NUMBER_OF_DEMO_PROTOCOLS] = 
   {"Z3950V2","Z39501988"};

  /* Register the client application-layer protocols. */
  appConnClientInit();

  for(j=0; j<NUMBER_OF_DEMO_PROTOCOLS; j++)
   {
     nConn* net = NULL;
     appConn* conn = NULL;
     list diags = NULL;

     printf("\n\n\n*****Demonstrating application protocol %s.*****\n",
	    appConnTypeNameList[j]);

     /* Open a TCP connection. */
     if (nConnOpen(&net,nConnTypeName,2L,host,port) != nConnStatusOK)
      {
	fprintf(stderr,
		"Unable to open a %s network connection to %s port %ld.\n",
		nConnTypeName,host,port);
	nConnTypesFree();
	appConnTypesFree();
	exit(1);
      }

     /* Using the TCP connection, open a application protocol connection. */
     if (appConnOpen(&conn,appConnTypeNameList[j],net,buf,BUFFER_SIZE,1L,0L)
	 != appConnStatusOK)
      {
	fprintf(stderr,"Unable to open a %s application connection.\n",
	       appConnTypeNameList[j]);
	nConnTypesFree();
	appConnTypesFree();
	nConnFree(net);
	exit(1);
      }

     /* Initialize and display any diagnostics that result. */
     if (!init(conn,TEST_CLIENT,&diags))
      {
	fprintf(stderr,
		"Unable to initialize %s connection using protocol %s.\n",
		nConnTypeName,appConnTypeNameList[j]);
	displayDiagsAndCleanup(NULL,NULL,NULL,diags);
	nConnTypesFree();
	appConnTypesFree();
	nConnFree(net);
	appConnFree(conn);
	exit(1);
      }

     /* Run the search and retrieval demonstration tests on the 
	tcp network connection, using the application level protocol. */
     testSearchAndRetrieveByRecId(conn);
     testSearchWithRFAndRetrieveByPosition(conn);

     /* Clean up the TCP nConn connection. */
     nConnFree(net);

     /* Clean up the Application appConn connection. */
     appConnFree(conn);
   }

  printf("\n\n\n*****Demonstration Complete*****\n");
  
  nConnTypesFree();
  appConnTypesFree();
}






