/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Module:  appConnCli.h

 Purpose: This header file contains the client Application Programmer
          Interface (API).  It contains client-specific API functions.

 Notes:   Transactions require an application connection handle of type 
          appConn, and return an explanatory list of diagnostics and a
          status code.  Each diagnostic in the diagnostic list is a handle
          to a structure of type appConnDiag, which contains a diagnostic
          code of type appConnDiagCode and an explanatory string.  The 
          status code is of type appConnStatus.  The diagnostic structure
          and codes, and the status codes can be found in the file 
          "appConnMgr.h".  

          Memory management for all inputs and outputs must be handled
          by the caller; i.e. the caller must free all inputs and outputs.
          Unless otherwise noted, inputs are not modified by the callee.

          Only one database access per connection.  Multiple simultaneous
          connections may be used.

          All character strings are null terminated.
----------------------------------------------------------------------------*/

#ifndef appConnCli_h
#define appConnCli_h

#include "appConnMgr.h"



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnClientInit
 Purpose:     Initializes the API.  This function must be called before using
              the client API functions defined in this file.
 Called by:   Called by client program.
 Parameters:  None.
 Returns:     void
----------------------------------------------------------------------------*/
void appConnClientInit(void);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnInitReqWrite
 Purpose:     Writes an initialization request packet to the connection.
 Called by:   Called by client program.
 Parameters:  conn       -- input; a handle to an appConn. It holds
                            information about the state of the connection; 
                            may be modified by this function.
              user       -- input; a null-terminated character string 
                            containing the name of the user initiatiating
                            the transaction.
              password   -- input; a null-terminated character string
                            containing the client password.
              clientInfo -- input; a null-terminated character string 
                            containing additional client information to be
                            sent to the server.
              userInfo   -- input; unused.
              diags      -- output; a handle to a list of diagnostics,
                            where each item in the list is of type appConnDiag
                            (see appConnMgr.c).  If the operation was 
                            unsuccessful, a list of diags are returned in
                            this parameter.
 Returns:     appConnStatus
 Imp Notes:   In Z39.50, initialization is used for user authentication and
              to negotiate the transmitted buffer size.  In the future, it
              may be used to negotiate other parameters, such as preferred
              language or character set.
              In this release, id authentication (user, password, and
              clientInfo) are ignored.  In the future, it is expected that
              user authentication will be implemented.
----------------------------------------------------------------------------*/
appConnStatus appConnInitReqWrite(appConn* conn,
                                  char*    user,    
                                  char*    password,
                                  char*    clientInfo,
                                  any*     userInfo,
				  list*    diags);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnInitRespRead
 Purpose:     Reads an initialization response packet from the connection.
              If the connection is refused, a list of explanatory diagnostics
              is returned in diags.
 Called by:   Called by client program.
 Parameters:  conn       -- input; a handle to an appConn. It holds
                            information about the state of the connection; 
                            may be modified by this function.
              userInfo   -- output; used to hold additional information
                            returned from server.
              diags      -- output; a handle to a list of diagnostics,
                            where each item in the list is of type appConnDiag
                            (see appConnMgr.c).  If the operation was 
                            unsuccessful, a list of diags are returned in
                            this parameter.
 Returns:     appConnStatus
 Imp Notes:   The negotiated buffer size field of the connection is modified.
----------------------------------------------------------------------------*/
appConnStatus appConnInitRespRead(appConn* conn,
                                  any**    userInfo,
				  list*    diags);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnSearchReqWrite
 Purpose:     Writes a search request packet to the connection.
 Called by:   Called by client program.
 Parameters:  conn          -- input; a handle to an appConn. It holds
                               information about the state of the connection; 
                               may be modified by this function.
              maxHits       -- input; a long integer whose value is the
                               maximum number of results the server is 
                               requested to generate.
              hitsRequested -- input; a long integer.  The top-ranking
                               hitsRequested hits are returned to the client
                               in the call to appConnSearchRespRead().  More
                               may be retrieved by calls to 
                               appConNextHeadlinesReqWrite() and
                               appConNextHeadlinesRespRead().
              rankingScheme -- input; unused.
              dbName        -- input; a null-terminated string containing
                               the name of the database to search.
              qry           -- input; a handle to an appConnQry data type
                               containing the user's query and is created
                               using the appQryAddTerm() function.
              userInfo      -- input; unused.
              diags         -- output; a handle to a list of diagnostics,
                               where each item in the list is of type 
                               appConnDiag (see appConnMgr.c).  If the
                               operation was unsuccessful, a list of diags
                               are returned in this parameter.
 Returns:     appConnStatus
 Imp Notes:   In Z39.50-1988, the maximum result set size is equal to maxHits,
              and hitsRequested is ignored.  maxHits hits must fit into the
              negotiated buffer size.  The server does not maintain a result
              set and thus, the next set of headlines cannot be obtained.
              In Z39.50-V2, maxHits is the largeSetLowerBound, and the
              hitsRequested is the mediumSetPresentNumber.  In this release,
              the element set name and record syntax requested is "VARIANT" 
              and Generic Record Syntax, respectively, as specified in 
              the WAIS Profile of Z39.50-V2.  In a future release, other
              element sets and record syntaxes may be attempted in the
              event this initial request fails.
----------------------------------------------------------------------------*/
appConnStatus appConnSearchReqWrite(appConn*     conn,  
                                    long         maxHits,
                                    long         hitsRequested, 
                                    long         rankingScheme, 
                                    char*        dbName, 
                                    appConnQry*  qry,
                                    any*         userInfo,
				    list*        diags);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnSearchRespRead
 Purpose:     Reads a search response packet from the connection.  This 
              function reads data from the connection and parses it into
              hits.  The total number of hits in the result set is returned
              in the parameter totalHits; this value may be greater than or
              equal to the actual number of hits returned (i.e. the length
              of the hits list).
 Called by:   Called by client program.
 Parameters:  conn         -- input; a handle to an appConn. It holds
                              information about the state of the connection; 
                              may be modified by this function.
              totalHits    -- output; a pointer to an long integer whose value 
                              is the total number of hits in the result
                              set generated by the server.  The value
                              of totalHits is less than or equal to the
                              maxHits parameter specified in the call to
                              appConnSearchReqWrite().
              qryTermsUsed -- output; a handle to a list of null-terminated
                              character strings.  Each character string
                              contains the query terms used by the server
                              in the search; these terms may or may not
                              have been specified in the original query.
                              The terms are dependent on the behavior of the
                              server and the contents of the database.  
                              Stop words may have been eliminated, right-
                              truncated words may have been expanded, and
                              synonyms may have been included.
              hits         -- output; a handle to a list.  Each item
                              in the list is of type appConnHeadline, and
                              represents one element of the result set.  
                              The number of hits returned may be less than
                              or equal to the number of hits requested.
              userInfo     -- output; unused.
              diags        -- output; a handle to a list of diagnostics,
                              where each item in the list is of type 
                              appConnDiag (see appConnMgr.c).  If the
                              operation was unsuccessful, a list of diags
                              are returned in this parameter.
 Returns:     appConnStatus
 Imp Notes:   In Z39.50-V2, if the server does not support piggy-backed 
              presents, the number of hits returned may be zero, even though
              the total hits is non-zero.  To obtain the hits, use the
              appConnNextHeadlinesReqWrite and appConnNextHeadlinesRespRead
              API functions.
              In Z39.50-1988 in contrast, all hits that can be returned will
              be returned by this function.  No result set is stored on the
              server.
----------------------------------------------------------------------------*/
appConnStatus appConnSearchRespRead(appConn* conn,
                                    long*    totalHits, 
				    list*    qryTermsUsed,
                                    list*    hits,
                                    any**    userInfo,
				    list*    diags);
   


/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnNextHeadlinesReqWrite
 Purpose:     Writes a next headlines request packet to the connection.  This
              function requests more hits from the server.  Hits beginning at
              (startingFromHit) to (startingFromHit + hitsRequested - 1), 
              inclusive, are requested.
 Called by:   Called by client program.
 Parameters:  conn            -- input; a handle to an appConn. It holds
                                 information about the state of the 
                                 connection; may be modified by this function.
              startingFromHit -- input; a long integer whose value is
                                 the starting position of the hits in the
                                 result list that are being requested.
              hitsRequested   -- input; a long integer whose value is 
                                 the number of hits being requested.
              userInfo        -- input; unused.
              diags           -- output; a handle to a list of diagnostics,
                                 where each item in the list is of type 
                                 appConnDiag (see appConnMgr.c).  If the
                                 operation was unsuccessful, a list of diags
                                 are returned in this parameter.
 Returns:     appConnStatus
 Imp Notes:   This function is undefined for Z39.50-1988.  An appConnStatus
              of appConnStatusUndefinedFunction is returned.
              In Z39.50-V2, this function is implemented as a present, using
              the same element set name and record syntax as for the search
              request.
----------------------------------------------------------------------------*/
appConnStatus appConnNextHeadlinesReqWrite(appConn* conn,
                                           long     startingFromHit, 
				           long     hitsRequested,
				           any*     userInfo,
					   list*    diags);
 




/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnNextHeadlinesRespRead
 Purpose:     Reads a next headlines response packet from the connection.
 Called by:   Called by client program.
 Parameters:  conn       -- input; a handle to an appConn. It holds
                            information about the state of the connection;
                            may be modified by this function.
              hits       -- output; a handle to a list where each item
                            in the list is of type appConnHeadline, and
                            represents one element of the requested result
                            set.  The number of hits returned may be less
                            than or equal to the number of hits requested.
              userInfo   -- output; unused.
              diags      -- output; a handle to a list of diagnostics,
                            where each item in the list is of type 
                            appConnDiag (see appConnMgr.c).  If the
                            operation was unsuccessful, a list of diags
                            are returned in this parameter.
 Returns:     appConnStatus
 Imp Notes:   This function is undefined for Z39.50-1988.  An appConnStatus
              of appConnStatusUndefinedFunction is returned.
              In Z39.50-V2, this function is implemented as a present.
----------------------------------------------------------------------------*/
appConnStatus appConnNextHeadlinesRespRead(appConn* conn,
					   list*    hits,
					   any**    userInfo,
					   list*    diags);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnRetrieveReqWrite
 Purpose:     Writes a retrieval request packet to the connection.  Retrieval
              may be performed using one of three methods.  The first method
              of retrieval is to select a hit from a previous search; this
              retrieval is specified by the hit's position in the result list.
              The second method is to specify the document identifier of the
              requested document, and the third is to specify the URx (Uniform
              Resource Locator or Name) of the requested document.  In all 
              cases, if the document is composed of multiple components,
              or elements, (such as an citation, an abstract, a table of 
              contents, chapters, sections, etc.), an element specifier may
              be specified.  If an element is available in multiple variants
              (such as different languages, formats, etc.), a variant 
              identifier may also be specified.  If no element identifier or
              variant identifier is specified, a default element and variant
              is selected by the server.
 Called by:   Called by client program.
 Parameters:  conn          -- input; a handle to an appConn. It holds
                               information about the state of the connection;
                               may be modified by this function.
              retrievalType -- input;; an long integer used to specify the 
                               retrieval method.  Must be one of the following
                               defined in the file "appConnMgr.h":
                                   appConnRetrieveByURx
                                   appConnRetrieveByRecId
                                   appConnRetrieveByPosition
              resSetPos     -- input; a long integer used to specify the 
                               position in the result set of the document
                               that is requested.  Applicable only for the
                               retrieval type of appConnRetrieveByPosition.
              recId         -- input; a handle to an 'any' datatype;
                               specifies the record identifier of the
                               requested document.
              docId         -- input; a null-terminated character string;
                               specifies the record identifier or URx
                               of the requested document.
              elemId        -- input; a null-terminated character string;
                               specifies the element identifier of the
                               requested document.
              varId         -- input; a null-terminated character string; 
                               specifies the variant identifier of the
                               requested document.
              dbName        -- input; a null-terminated string.  It
                               contains the name of the database from which
                               to retrieve the data.  Only required when
                               the retrieval type is appConnRetrieveByUrx
                               or appConnRetrieveByRecId.
              startByte     -- input; a long integer used to specify the
                               starting byte of the data to be returned.
              lengthInBytes -- input; a long integer used to specify how 
                               many bytes of data are to be returned.
              userInfo      -- input; unused.
              diags         -- output; a handle to a list of diagnostics,
                               where each item in the list is of type 
                               appConnDiag (see appConnMgr.c).  If the
                               operation was unsuccessful, a list of diags
                               are returned in this parameter.
 Returns:     appConnStatus
 Imp Notes:  In Z39.50-1988, retrieval is implemented using a search.  Thus,
             only retrieval by RecId is possible, using a retrieval type 
             appConnRetrieveByRecId.
             In Z39.50-V2, all three retrieval types are possible.  Retrieval
             by position, using retrieval type appConnRetrieveByPosition, is 
             implemented as a present.  Retrieval by URx and RecId are
             implemented as a search.
             In Z39.50-V2, startByte and lengthInBytes are ignored.  The
             first retrieval returns the first piece of the document, fitting
             as much in the packet as can fit.  Successive calls to retrieve
             the same document return successive pieces of the document.
             Each successive piece begins where the last piece left off.
----------------------------------------------------------------------------*/
appConnStatus appConnRetrieveReqWrite(appConn*         conn,
	 			      long             retrievalType,
				      long             resSetPos,
				      any*             recId,
				      char*            docId,
				      char*            elemID,
				      char*            varID,
				      char*            dbName,
				      long             startByte,
				      long             lengthInBytes,
				      any*             userInfo,
				      list*            diags);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnRetrieveRespRead
 Purpose:     Reads a retrieval response packet from the connection.  A 
              diagnostic error code of appConnDiagRequestOutOfRange is 
              returned in the list of diagnostics if the data returned
              contains the end of the document.
 Called by:   Called by client program.
 Parameters:  conn       -- input; a handle to an appConn. It holds
                            information about the state of the connection;
                            may be modified by this function.
              data       -- output; a pointer to a null-terminated character
                            string containing the requested data.
              dataLength -- output; a pointer to a long integer whose value
                            contains the number of bytes in data.
              userInfo   -- output; unused.
              diags      -- output; a handle to a list of diagnostics,
                            where each item in the list is of type 
                            appConnDiag (see appConnMgr.c).  If the
                            operation was unsuccessful, a list of diags
                            are returned in this parameter.
 Returns:     appConnStatus
----------------------------------------------------------------------------*/
appConnStatus appConnRetrieveRespRead(appConn* conn,
				      char**   data,
				      long*    dataLength,
				      any**    userInfo,
				      list*    diags);


#endif /* ndef appConnCli_h */
