/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Module:   appConnMgr.h

 Purpose:  This file defines an abstract way of communicating with a remote 
           client or server.  It contains a library of functions that may be
           used for both the client and the server.  The API is flexible 
           enough to support several different protocols.  This is done by
           defining a data structure called an appConn which is an abstract
           protocol connection.  
----------------------------------------------------------------------------*/

#ifndef appConnMgr_h
#define appConnMgr_h

#include "nConnMgr.h"
#include "cutil.h"
#include "list.h"
#include "date.h"
#include "waislog.h"


/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 DataType:  appConnStatus
 Purpose:   The appConnStatus data type is used for status returned from the
            client and server application API functions.  The corresponding
            constants are the possible values that the appConnStatus data
            type may assume.
----------------------------------------------------------------------------*/
typedef long appConnStatus;

#define appConnStatusUndefinedConnectionType (-1L)
#define appConnStatusOK                       (0L)
#define appConnStatusUndefinedFunction        (1L)
#define appConnStatusBadArguments             (2L)
#define appConnStatusNotFinished              (3L)

/*----------------------------------------------------------------------------
 Function:    appConnStatusLogStatus
 Purpose:     Log status to a log file.
 Called by:   Called by client or server program.
 Parameters:  status -- input; of type appConnStatus.
 Returns:     void
----------------------------------------------------------------------------*/
void appConnStatusLogStatus(appConnStatus status);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Query Types: A query may consist of a single term or set of terms separated 
              by Boolean operators.  A query is represented as a binary tree, 
              where the leaves of the tree are each a node tagged with one of
              the following types (constants) defined below:
 Imp Notes:   In Z39.50, each term may be associated with one of several 
              attribute types, such as use, structure, relation, completeness, 
              etc.  For this release, indivicual attributes types for each
              term are not enumerated, but instead are dealt with in the
              lower-level implmentation-specific peices of the protocol.
 ----------------------------------------------------------------------------*/
#define qryLeafUndefined (-1L) /* An undefined query */
#define qryLeafFreeText   (0L) /* Free-Form Human-Entered Text. */
#define qryLeafRecId      (1L) /* Local Record Identifier */
#define qryLeafDocId      (2L) /* Universal Record Identifier */
#define qryLeafRFDocText  (3L) /* Relevance Feedback by Document Text. */
#define qryLeafRFRecId    (4L) /* Relevance Feedback by document Key */
#define qryLeafRFDocId    (5L) /* Relevance Feedback by URx */
#define qrySubTree        (6L) /* A query sub tree (implements parens) */

/*----------------------------------------------------------------------------
 Boolean Op:  Each non-leaf node of the query is made up of a Boolean operator
              tagged with one of the following types (constants) defined
              below:
 ----------------------------------------------------------------------------*/
#define qryNodeAnd     (0L)
#define qryNodeOr      (1L)
#define qryNodeAndNot  (2L)

/*----------------------------------------------------------------------------
 DataType: appConnQryRF
 Purpose:  The appConnQryRF data type is used to a represent relevant piece
           of a document that is to be used in a query for relevance feedback.
 ----------------------------------------------------------------------------*/
typedef struct appConnQryRF
{ any*    recId;
  char*   docId;
  char*   elemId;
  char*   varId;
  long    startByte;
  long    lengthInBytes;
} appConnQryRF;

/*----------------------------------------------------------------------------
 DataType: appConnQry
 Purpose:  The appConnQry data type is used for to represent the text typed
           by the user, or the text of a relevant piece of a document.
           An instance of appConnQry is used as the leaf node in the query.
 ----------------------------------------------------------------------------*/
typedef struct appConnQry
{ long                type;
  long                operator;
  any*                term;
  appConnQryRF*       rf;
  struct appConnQry*  arg1;  
  struct appConnQry*  arg2;
} appConnQry;

/*----------------------------------------------------------------------------
 Function:    appConnQryMakeLeafFreeText
 Purpose:     Create and return a leaf node query of free text.
 Called by:   Called by client program.
 Parameters:  freeText -- input; a null-terminated character string containing
                          free text typed by the user.
 Returns:     Returns a handle to the created query.
 ----------------------------------------------------------------------------*/
appConnQry* appConnQryMakeLeafFreeText(char* freeText);

/*----------------------------------------------------------------------------
 Function:    appConnQryMakeLeafRFDocText
 Purpose:     Create and return a leaf node query containing the text of
              a relevant document.
 Called by:   Called by client program.
 Parameters:  rfText -- input; a null-terminated character string containing
                        text taken from a document to be used in relevance
                        feedback.
 Returns:     Returns a handle to the created query.
 ----------------------------------------------------------------------------*/
appConnQry* appConnQryMakeLeafRFDocText(char* rfText);

/*----------------------------------------------------------------------------
 Function:    appConnQryMakeLeafRFRecId
 Purpose:     Create and return a leaf node query containing the local record
              identifier of a relevant document.
 Called by:   Called by client program.
 Parameters:  See parameters to the client API function,
              appConnRetrieveReqWrite() in appConnCli.h
              The parameters are not copied.
 Returns:     Returns a handle to the created query.
 ----------------------------------------------------------------------------*/
appConnQry* appConnQryMakeLeafRFRecId(any*    recId,
				      char*   elemId,
				      char*   varId,
				      long    startByte,
				      long    lengthInBytes);

/*----------------------------------------------------------------------------
 Function:    appConnQryMakeLeafRFDocId
 Purpose:     Create and return a leaf node query containing the URx of a
              relevant document.
 Called by:   Called by client program.
 Parameters:  See parameters to the client API function,
              appConnRetrieveReqWrite() in appConnCli.h
              The parameters are not copied.
 Returns:     Returns a handle to the created query.
 ----------------------------------------------------------------------------*/
appConnQry* appConnQryMakeLeafRFDocId(char*   docId,
				      char*   elemId,
				      char*   varId,
				      long    startByte,
				      long    lengthInBytes);

/*----------------------------------------------------------------------------
 Function:    appConnQryMakeNode
 Purpose:     Create an return an intermediate node of a query.  The
              intermediate node represents a Boolean operation on two queries.
 Called by:   Called by the client program.
 Parameters:  operator   -- input; a long integer whose value is one of the 
                            constants: qryNodeAnd, qryNodeOr, or 
                            qryNodeAndNot.
              arg1, arg2 -- input; an intermediate or leaf node query of type
                            appConnQry.
 Returns:     Returns a handle to the created query.
 ----------------------------------------------------------------------------*/
appConnQry* appConnQryMakeNode(long        operator,
			       appConnQry* arg1,
			       appConnQry* arg2);  

/*----------------------------------------------------------------------------
 Function:    appConnQryFree
 Purpose:     Frees an appConnQry and any sub qry's.
 Called by:   Called by client or server program.
 Parameters:  qry -- input; a pointer to query of type appConnQry.
 Returns:     void
 ----------------------------------------------------------------------------*/
void appConnQryFree(appConnQry* qry);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Datatype: appConn
 Purpose:  The appConn data type is the default data structure used to contain
           information on a specific instance of a client/server connection.
           The slots in this datastructure are be transparent to the user 
           of the API functions.
 ----------------------------------------------------------------------------*/
typedef struct appConn
{ char*   name;
  nConn*  net;
  char*   buf; 
  long    realBufSz;
  long    negotiatedBufSz;
  long    timeOutSecs;
} appConn;

/*----------------------------------------------------------------------------
 Function:    appConnOpen
 Purpose:     Creates an appConn appropriate for the specified protocol.
              The variable arguments are used to allow any implementation
              specific values to be passed in by a knowledgable client.
 Called by:   Called by client or server program.
 Parameters:  conn        -- output; a pointer to the handle of the
                             appConn. It holds information about the state
                             of the connection; created by this function.
              name        -- input; a null-terminated character string for
                             the name of the protocol to be used.  For this
                             release, name must be one of "Z39501988" or
                             "Z3950V2".
              net         -- input; a handle to the network connection nconn.
                             The network connection is created by the 
                             nConnOpen() function (see nConnMgr.h).
              buffer      -- input; a character buffer used to hold the 
                             incoming and outgoing data to be transmitted.
                             It is reused over the lifetime of the connection.
                             Must be created by caller.
              bufSz       -- input; a long integer whose value is the size of
                             the buffer.
              timeOutSecs -- input; a long integer whose value is the timeout
                             time for a transaction.
              numArgs     -- input; a long integer for the number of 
                             additional arguments to this function.
 Returns:     void

 ----------------------------------------------------------------------------*/
appConnStatus appConnOpen(appConn** conn,
			  char*     name,
			  nConn*    net,
			  char*     buffer,
			  long      bufSz,
			  long      timeOutSecs,
			  long      numArgs,
			  ...);

/*----------------------------------------------------------------------------
 Function:    appConnFree
 Purpose:     Frees the appConn, and any memory related to the 
              implementation-specific pieces of the protocol.  
              Note that this does not free the network connection nConn.
 Called by:   Called by client or server program.
 Parameters:  conn -- input; a handle to an appConn. It holds information
                      about the state of the connection; it is freed by this
                      function.
 Returns:     appConnStatus

 ----------------------------------------------------------------------------*/
appConnStatus appConnFree(appConn* conn);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Datatype: appConnType
 Purpose:  The appConnType data type contains an extensible table of function
           pointers in the API.  These include both the client and the server
           API functions, and there is one table for each protocol supported.
 ----------------------------------------------------------------------------*/
typedef struct appConnType
{ char* name;
  appConnStatus (*appConnOpenFunc)(appConn** conn,
				   char*     name,
				   nConn*    net,
				   char*     buffer,
				   long      bufSz,
				   long      timeOutSecs,
				   long      numArgs,
				   ...);
  appConnStatus (*appConnFreeFunc)(appConn* conn);
  appConnStatus (*pduInfoFunc)(appConn* conn,
			       long*    pduType,
			       long*    packetLength,
			       list*    diags);
  appConnStatus (*initReqWriteFunc)(appConn* conn,
				    char*    user,    
				    char*    password,
				    char*    clientInfo,
				    any*     userInfo,
				    list*    diags);
  appConnStatus (*initReqReadFunc)(appConn* conn,
				   char**   user,    
				   char**   password,
				   char**   clientInfo,
				   any**    userInfo,
				   list*    diags);
  appConnStatus (*initRespWriteFunc)(appConn* conn,  
				     boolean  connectionAccepted,
				     any*     userInfo,
				     list*    diags);
  appConnStatus (*initRespReadFunc)(appConn* conn,  
				    any**    userInfo,
				    list*    diags);
  appConnStatus (*searchReqWriteFunc)(appConn*     conn,  
				      long         maxHits,
				      long         hitsRequested, 
				      long         rankingScheme, 
				      char*        dbName, 
				      appConnQry*  qry,
				      any*         userInfo,
				      list*        diags);
  appConnStatus (*searchReqReadFunc)(appConn*     conn,  
				     long*        maxHits,
				     long*        hitsRequested, 
				     long*        rankingScheme, 
				     char**       dbName, 
				     appConnQry** qry,
				     any**        userInfo,
				     list*        diags);
  appConnStatus (*searchRespWriteFunc)(appConn* conn,
				       long     totalHits, 
				       list     qryTermsUsed,
				       list     hits,
				       list     userDiags,
				       any*     userInfo,
				       list*    diags);
  appConnStatus (*searchRespReadFunc)(appConn* conn,
				      long*    totalHits, 
				      list*    qryTermsUsed,
				      list*    hits,
				      any**    userInfo,
				      list*    diags);
  appConnStatus (*nextHeadlinesReqWriteFunc)(appConn* conn,
					     long     startingFromHit, 
					     long     hitsRequested,
					     any*     userInfo,
					     list*    diags);
  appConnStatus (*nextHeadlinesReqReadFunc)(appConn* conn,
					    long*    startingFromHit, 
					    long*    hitsRequested,
					    any**    userInfo,
					    list*    diags);
  appConnStatus (*nextHeadlinesRespWriteFunc)(appConn* conn,
					      list     hits,
					      list     userDiags,
					      any*     userInfo,
					      list*    diags);
  appConnStatus (*nextHeadlinesRespReadFunc)(appConn* conn,
					     list*    hits,
					     any**    userInfo,
					     list*    diags);
  appConnStatus (*retrieveReqWriteFunc)(appConn* conn,
					long     retrievalType,
					long     resSetPos, 
					any*     recId, 
					char*    docId, 
					char*    elemId, 
					char*    varId, 
					char*    dbName,
					long     startByte,
					long     lengthInBytes,
					any*     userInfo,
					list*    diags);
  appConnStatus (*retrieveReqReadFunc)(appConn* conn,
				       long*    retrievalType,
				       long*    resSetPos,    
				       any**    recId,
				       char**   docId,
				       char**   elemId,        
				       char**   varId,        
				       char**   dbName,
				       long*    startByte,
				       long*    lengthInBytes,
				       any**    userInfo,
				       list*    diags);
  appConnStatus (*retrieveRespWriteFunc)(appConn* conn,  
					 char*    data,
					 long     dataLength,
					 list     userDiags,
					 any*     userInfo,
					 list*    diags);
  appConnStatus (*retrieveRespReadFunc)(appConn* conn, 
					char**   data,
					long*    dataLength,
					any**    userInfo,
					list*    diags);
} appConnType;

/*----------------------------------------------------------------------------
 Function:    registerAppConnType
 Purpose:     Registers the protocol (appConnType), if a protocol by the 
              same name has not been registered before,
 Called by:   initAppConnClient, initAppConnServer
 Globals:     appConns -- input; the global list of registered types is
                          updated with the new appConnType.  This variable
                          is modified by the function.
 Parameters:  type -- input; a handle to an appConnType containing the
                      functions to be used in the protocol.
 Returns:     void
 Imp Notes:   With this release, two protocols are registered by the name of
              "Z39501988" and "Z3950V2".  Others could be added in the future.
----------------------------------------------------------------------------*/
void registerAppConnType(appConnType* type);

/*----------------------------------------------------------------------------
 Function:    findAppConnTypeNamed
 Purpose:     Finds and returns the appconn with the given name.  If an appConn
              of that name is not found, NULL is returned.
 Called by:   Automatically called by all API functions in the appConnType
              data structure.  Not useful to the client or server program.
 Globals:     appConns -- input; uses the global list of appConns to determine
                          which implementation-specific API function to call.
 Parameters:  name     -- input; a null terminated character string containing
                          the name of a registered appConn.
 Returns:     Returns a handle to the appConnType if found, otherwise 
              returns NULL.
 ----------------------------------------------------------------------------*/
appConnType* findAppConnTypeNamed(char* name);

/*----------------------------------------------------------------------------
 Function:    freeAppConnType
 Purpose:     Frees the memory associated with the registered appConnTypes.
 Called by:   Called by client and server programs.
 Parameters:  none
 Returns:     void
 ----------------------------------------------------------------------------*/
void freeAppConnTypes(void);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Datatype:  appConnDiagCode
 Purpose:   The appConnDiagCode data type is used as a code for the diagnostic
            conditions returned in the diags list of the API functions. The 
            corresponding constants are the possible values that the 
            appConnDiagCode data type may assume.
 ----------------------------------------------------------------------------*/
typedef long appConnDiagCode;

#define appConnDiagConnErrOpen           (0L)
#define appConnDiagConnErrRead           (1L)
#define appConnDiagConnErrWrite          (2L)
#define appConnDiagConnClosed            (3L)
#define appConnDiagServerError           (4L)
#define appConnDiagProtocolError         (5L)
#define appConnDiagClientError           (6L)
#define appConnDiagAccessDenied          (7L)
#define appConnDiagUnsupportedSearch     (8L)
#define appConnDiagBadPacket             (9L)
#define appConnDiagNoMoreHeadlines      (10L)
#define appConnDiagNoDatabaseSpecified  (11L)
#define appConnDiagNoDocIdSpecified     (12L)
#define appConnDiagNoResultSetAvailable (13L)
#define appConnDiagDocumentNotFound     (14L)
#define appConnDiagInvalidChunkType     (15L)
#define appConnDiagInvalidDatabase      (16L)
#define appConnDiagInvalidDocId         (17L)
#define appConnDiagRequestOutOfRange    (18L)
#define appConnDiagNegotiatedBufTooSmall (19L)
#define appConnDiagNumber               (20L) /* update when adding new diag */

typedef struct diagTableEntry
{ long  priority;
  long  logType;
  char* message;
} diagTableEntry;

/*----------------------------------------------------------------------------
 Datatype:  appConnDiag
 Purpose:   The appConnDiag data type is used to hold the diagnostic error
            code and a string containing a description of the diagnostic.
            MAX_DIAG_INFO_LEN is the maximum size of the diag info string.
 ----------------------------------------------------------------------------*/
#define MAX_DIAG_INFO_LEN 1000L

typedef struct appConnDiag
{ appConnDiagCode code;
  char*           info;
} appConnDiag;

/*----------------------------------------------------------------------------
 Function:    appConnDiagMake
 Purpose:     Create and return a diagnostic
 Called by:   Called by server program.
 Parameters:  code       -- input; an appConnDiag error.
              infoFormat -- input; a string that may contain format conversion
                            specifications for the additional optional
                            arguments (similar to printf).
 Returns:     A pointer to the new data structure of type appConnDiag.
----------------------------------------------------------------------------*/
appConnDiag* appConnDiagMake(appConnDiagCode code,
			     char*           infoFormat,
			     ...);

/*----------------------------------------------------------------------------
 Function:    appConnDiagLogErrors
 Purpose:     Free the diagnostic data structure.
 Called by:   Called by client or server program.
 Parameters:  diag -- input; an appConnDiag to be freed.
 Returns:     void
----------------------------------------------------------------------------*/
void appConnDiagFree(appConnDiag* diag);

/*----------------------------------------------------------------------------
 Function:    appConnDiagLogErrors
 Purpose:     Log the diagnostics to a log file.
 Called by:   Called by client or server program.
 Parameters:  diags -- input; a list of diags of the datatype appConnDiag.
 Returns:     void
----------------------------------------------------------------------------*/
void appConnDiagLogErrors(list diags);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Datatype: appConnVariant
 Purpose:  The appConnVariant data structure is used to choose among 
           several representations of a retrievable object.  For example,
           a picture may be available in TIFF and GIF formats. each of
           those is a separate variant.
 ----------------------------------------------------------------------------*/
typedef struct appConnVariant
{ char*  name;         /* variant identifier */
  char*  composition;  /* Internet Media Type */
  char*  language;     /* Language code */
  char*  characterSet; /* Character set */
  long   size;         /* Size of variant in bytes */
  double cost;         /* Cost of the element */
  long   currency;     /* Currency of the cost, from ISO 4217 */
} appConnVariant;

/*----------------------------------------------------------------------------
 Function:    appConnMakeVariant
 Purpose:     Create a variant structure, and copy in the arguments.
 Called by:   Called by the server program.
 Parameters:  name         -- input; a null-terminated character string
                              containing an identifying name for the variant.
              composition  -- input; a null-terminated character string 
                              containing the media type (MIME type) for the
                              variant.
              language     -- input; a null-terminated character string
                              containing the language of the variant.
                              The language is specified by the USMARC code
                              list for languages.
              characterSet -- input; a null-terminated character string
                              containing the character set this variant is
                              available in.  The character set is specified
                              by the International Register of Coded
                              Character Sets.
              size         -- input; a long integer whose value contains
                              the total size of the variant in bytes.
              cost         -- input; a long integer whose value contains
                              the cost of retrieving the document.
              currency     -- input; a long integer for the currency of the
                              cost value.  The currency values are 
                              currently unspecified.
 Returns:     Returns a handle to the newly created appConnVariant.
 ----------------------------------------------------------------------------*/
appConnVariant* appConnMakeVariant(char* name,
				   char* composition,
				   char* language,
				   char* characterSet,
				   long  size,
				   long  cost,
				   long  currency);

/*----------------------------------------------------------------------------
 Function:    appConnVariantFree
 Purpose:     Frees a variant and all its subfields.
 Called by:   appConnVariantListFree
 Parameters:  variant -- input; a handle for the appConnVariant to be freed.
 Returns:     void
 ----------------------------------------------------------------------------*/
void appConnVariantFree(appConnVariant* variant);

/*----------------------------------------------------------------------------
 Function:    appConnVariantListFree
 Purpose:     Free each variant in the list and the list itself.
 Called by:   appConnElementFree
 Parameters:  variants -- input; a list where each item in the list is an 
                          appConnVariant data type.
 Returns:     void
 ----------------------------------------------------------------------------*/
void appConnVariantListFree(list variants);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Datatype: appConnElement
 Purpose:  The appConnElement data type describes subcomponents of a 
           retrievable object, for example a newspaper article or an image.
           A headline may have several elements (eg. an article and an image).
           Also, elements may have sub elements (eg. a book and its chapters).
 ----------------------------------------------------------------------------*/
typedef struct appConnElement
{ char* name;        /* A displayable name for this element */
  char* usageRight;  /* [optional] Text for display to user. Do we want this?*/
  list  variants;    /* Variants of this element */
  list  subElements; /* For elements with substructure */
} appConnElement;    /* Only leaf-level elements have variants */

/*----------------------------------------------------------------------------
 Function:    appConnMakeElement
 Purpose:     Creates an appConnElement structure, and copies in the name
              and usageRight.  It does not copy the variants or subElements.
 Called by:   Called by the server program.
 Parameters:  name        -- input; a null-terminated character string
                             containing a descriptive name identifying
                             the element.
              usageRight  -- input; a null-terminated character containing the
                             usage characteristics (such as copyright 
                             restrictions) of this element.
              variants    -- input; a list where each item in the list is of
                             type appConnVariant.  Gives the retreivable
                             variants for this element.
              subElements -- input; a list where each item in the list is
                             of type appConnElement.  Used for elements
                             containing structure, as a book may have chapters.
 Returns:     Returns a handle to the newly created appConnElement.
 ----------------------------------------------------------------------------*/
appConnElement* appConnMakeElement(char* name,
				   char* usageRight,
				   list  variants,
				   list  subElements);

/*----------------------------------------------------------------------------
 Function:    appConnElementFree
 Purpose:     Frees the element and all of its subvariants and subelements.
 Called by:   appConnElementListFree
 Parameters:  element  -- input; a handle for the appConnElement to be freed.
 Returns:     void
 ----------------------------------------------------------------------------*/
void appConnElementFree(appConnElement* element);

/*----------------------------------------------------------------------------
 Function:    appConnElementListFree
 Purpose:     Free each element in the list and the list itself.
 Called by:   appConnHeadlineFree
 Parameters:  elements -- input; a list where each item in the list is an 
                          appConnElement data type.
 Returns:     void
 ----------------------------------------------------------------------------*/
void appConnElementListFree(list elements);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Datatype: appConnHeadline
 Purpose:  
 ----------------------------------------------------------------------------*/
typedef struct appConnHeadline
{ char* headline;  /* A set of words that convey the main idea of the record */
  char* name;      /* [optional] An individual associated with the record.*/
  date* date;      /* [optional] The date associated with the record. */
  long  rank;      /* [optional] The rank ordering of the record. */
  long  score;     /* [optional] score of the record, 1-1000, inclusive. */
  any*  recId;     /* The unique identifier for the record. */
  char* urx;       /* A URN or URL associated with the record. */
  list  elements;  /* A list of the elements associated with this record. */
} appConnHeadline;

/*----------------------------------------------------------------------------
 Function:    appConnMakeHeadline
 Purpose:     Create and return an appConnHeadline.  All the parameters are
              copied except the element list.
 Called by:   Called by the server program.
 Parameters:  headline -- input; a null-terminated character string containing
                          a set of words that convey the main idea of the
                          document.
              name     -- input; a null-terminated character string consisting
                          of one or more individuals associated with one of the
                          elements; it could, for example, be an author or
                          an organization.  As another example, there might
                          be three elements: a fingerprint file, photo, and
                          a resume; the Name would be the individual that
                          these elements describe.
              date     -- input; a date structure associated with the record
                          (see date.h).
              rank     -- input; a long integer whose value is the position
                          of the headline in the result list.
              score    -- input; a long integer representing the numerical
                          score of the record based on its relevance to the
                          query, from 1 to 1000, inclusive.
              recId    -- input; a handle to an 'any' data type, that 
                          identifies the document, defined by, and unique
                          within the server system (it may, but need not be,
                          a URx).
              urx      -- input; a null-terminated character string containing
                          the Universal Resource Locator or Universal Resource
                          Name for document; used as input.
              elements -- input; a list of items where each item is of type 
                          appConnElement.  Each item in the list is a 
                          component of the document.
 Returns:     Returns a handle to the newly created appConnHeadline.
 ----------------------------------------------------------------------------*/
appConnHeadline* appConnMakeHeadline(char* headline,
				     char* name,
				     date* date,
				     long  rank,
				     long  score,
				     any*  recId,
				     char* urx,
				     list  elements);

/*----------------------------------------------------------------------------
 Function:    appConnHeadlineFree
 Purpose:     Free the appConnHeadline and all its components (including 
              the element and variant lists.
 Called by:   The client or server program.
 Parameters:  headline -- input; a handle for the appConnHeadline to be freed.
 Returns:     void
 ----------------------------------------------------------------------------*/
void appConnHeadlineFree(appConnHeadline* headline);



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Function:    appConnPduInfo
 Purpose:     The appConnPduInfo function is used primarily for dispatching.
              It reads the data on the connection and determines the pdu type
              and the length of the packet.  The data read is stored in the
              connection's buffer.
              Note that this function may return a status of 
              appConnStatusNotFinished, which indicates that not all of the
              data has yet been received and parsed.  The client should 
              continue to call the function until it returns either an error
              or non-error status.  This is designed to allow the client
              application to process other events and user input while
              waiting for the response to arrive.
 Called by:   Called by the client or server program.
 Parameters:  conn         -- input; a handle to an appConn. It holds
                              information about the state of the connection;
                              may be modified by this function.
              pduType      -- output; a handle to a long integer whose value
                              is one of the defined pdu types (listed below as 
                              constants).
              packetLength -- output; a handle to a long integer whose value
                              is the length of the received packet, or 
                              appConnUnknownPacketLength (-1) if the length is
                              unknown.
              diags        -- output; a handle to a list of diagnostics,
                              where each item in the list is of type 
                              appConnDiag (see appConnMgr.c).  If the
                              operation was unsuccessful, a list of diags
                              are returned in this parameter.
 Returns:     appConnstatus
 Imp Notes:   Internally, the function determines the pdu length.  
              In Z39.50-1988, the length is read from the wais header.
              In Z39.50-V2, the length is either read in the header,
              or it is read at the end of the packet (for indeterminate
              lengths).
 ----------------------------------------------------------------------------*/
#define appConnUnknown             (0L) 
#define appConnInitReq             (1L)
#define appConnInitResp            (2L)
#define appConnSearchReq           (3L)
#define appConnSearchResp          (4L)
#define appConnNextHeadlinesReq    (5L)
#define appConnNextHeadlinesResp   (6L)
#define appConnRetrieveReq         (7L)
#define appConnRetrieveResp        (8L)
#define appConnAccessControlReq    (9L)
#define appConnAccessControlResp  (10L)

#define appConnUnknownPacketLength (-1L)

appConnStatus appConnPduInfo(appConn* conn,
			     long*    pduType,
			     long*    packetLength,
			     list*    diags);


/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Constants: Ranking Schemes
 ----------------------------------------------------------------------------*/
#define rankByServerDefault (0L)
#define rankByScore         (1L)
#define rankByDate          (2L)
#define rankByHeadline      (3L)



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Constants: Retrieval Type
 ----------------------------------------------------------------------------*/
#define appConnRetrieveByURx       (0L)
#define appConnRetrieveByRecId     (1L)
#define appConnRetrieveByPosition  (2L)



/*----------------------------------------------------------------------------
 -----------------------------------------------------------------------------
 Function:    timeOut
 Purpose:     For use by the low-level implementation-specific functions
              to determine when to timeout.
 Called by:   Called by low-level implementation-specific functions to
              determine when to timeout.
 Parameters:  startTime   -- input; a time_t data type containing the time when
                             the operation was started.
              timeOutSecs -- input; a long integer for the amount of time in
                             secs to elapse from the startTime before timeout
                             is reached.
 Returns:     Returns a Boolean; returns 'true' the current time is past
              startTime + timeOutSecs, otherwise 'false' is returned.

 ----------------------------------------------------------------------------*/
boolean timeOut(time_t startTime,long timeOutSecs);



#endif /* ndef appConnMgr_h */
