/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Module:  appConnMgr.c

 Purpose: This file is the interface to the various application protocol
          abstractions.  The idea is to make it easy to add new protocols.

          To add a new application protocol type:
          1. Write the creation and API functions.  The creation function is
          of the form:

          appConnStatus appConnOpen(appCon** conn,char* name,nConn* net,
                                    char* buffer,long bufSz,long timeOutSecs,
		                    long numArgs,...)

          The function mallocs a structure which is based on an appConnn
          (ie. it may be an appConn or it may be a bigger structure whose
          first variable is an appConn) and assign its pointer to *conn.  
          Then it interprets its arguments using the function va_arg()
          defined in the ansi stdarg package.  These arguments are used
          to open a protocol connection to a remote system.  When the
          connection is established, or an error occurs, appConnStatus
          is returned.

          2. Include the header file for the new connection in the 
          appConnCli.c and appConnSrv.c files.

          3. In the appConnClientInit() and appConnServerInit() functions,
          add a call to register your new application protocol.
----------------------------------------------------------------------------*/

#include "appConnMgr.h"
#include "waislog.h"

/*-------------------------------------------------------------------------- */
/* Routines to manage the appConn protocols available */
/*-------------------------------------------------------------------------- */

static list appConns = NULL; /* a list of all registered protocols */

/*-------------------------------------------------------------------------- */

void 
registerAppConnType(appConnType* type)
{
  if (!findAppConnTypeNamed(type->name))
    appConns = collecting(appConns,type);
}

appConnType* 
findAppConnTypeNamed(char* name)
{
  long i;

  for (i = 0; i < length(appConns); i++)
   { appConnType* type = (appConnType*)nth(appConns,i);
     if (substrcmp(type->name,name))
       return(type);
   }
  return(NULL);
}

/*---------------------------------------------------------------------------*/

void
appConnTypesFree(void)
{
  appConnType* type;
  long i;

  for(i=0; i<length(appConns); i++)
   {
     type = nth(appConns,i);
     safeFree(type->name);
     safeFree(type);
   }

  freeList(appConns);
  appConns = NULL;
}

/*---------------------------------------------------------------------------*/

appConnStatus appConnFree(appConn* conn)
{
  appConnType* ai;

  if (conn == NULL)
    return(appConnStatusBadArguments);    

  if ((ai = findAppConnTypeNamed(conn->name)) == NULL ||
	   ai->appConnFreeFunc == NULL)
    return(appConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     safeFree(conn->name);
     res = (*ai->appConnFreeFunc)(conn);
     return(res);
   }
}

/*-------------------------------------------------------------------------- */

appConnStatus 
appConnOpen(appConn** conn,
	    char*     name,
	    nConn*    net,
	    char*     buffer,
	    long      bufSz,
	    long      timeOutSecs,
	    long      numArgs,
	    ...)
/* This routine creates an appConn appropriate for the specified name,
   using the network connection, buffer etc specified.  The variable 
   arguments are used to allow any implementation specific values to be 
   passed in by a knowledgable client.
 */
{
  va_list ap;	
  appConnStatus res;
  appConnType* ai = findAppConnTypeNamed(name);

  if (ai == NULL || ai->appConnOpenFunc == NULL || conn == NULL)
    return(appConnStatusUndefinedConnectionType);

  va_start(ap,numArgs);	
  res = (*ai->appConnOpenFunc)(conn,name,net,buffer,bufSz,
			       timeOutSecs,numArgs,ap);
  va_end(ap);	     

  return(res);
}

/*---------------------------------------------------------------------------*/
/* Queries */
/*---------------------------------------------------------------------------*/

appConnQry* 
appConnQryMakeLeafFreeText(char* freeText)
{
  appConnQry* qry;

  qry = (appConnQry*)safeMalloc(sizeof(appConnQry));
  qry->type = qryLeafFreeText;
  qry->term = stringToAny(freeText);

  return(qry);
}

/*---------------------------------------------------------------------------*/

appConnQry* 
appConnQryMakeLeafRFDocText(char* rfDocText)
{
  appConnQry* qry;

  qry = (appConnQry*)safeMalloc(sizeof(appConnQry));
  qry->type = qryLeafRFDocText;
  qry->term = stringToAny(rfDocText);

  return(qry);
}

/*---------------------------------------------------------------------------*/

appConnQry* 
appConnQryMakeLeafRFRecId(any*  recId,
			  char* elemId,
			  char* varId,
			  long  startByte,
			  long  lengthInBytes)
{
  appConnQry* qry;
  appConnQryRF* rfRecId;

  rfRecId = (appConnQryRF*)safeMalloc(sizeof(appConnQryRF));
  rfRecId->recId = duplicateAny(recId);
  rfRecId->elemId = safeStrdup(elemId);
  rfRecId->varId = safeStrdup(varId);
  rfRecId->startByte = startByte;
  rfRecId->lengthInBytes = lengthInBytes;

  qry = (appConnQry*)safeMalloc(sizeof(appConnQry));
  qry->type = qryLeafRFRecId;
  qry->term = duplicateAny(recId);
  qry->rf = rfRecId;

  return(qry);
}

/*---------------------------------------------------------------------------*/

appConnQry* 
appConnQryMakeLeafRFDocId(char* docId,
			  char* elemId,
			  char* varId,
			  long  startByte,
			  long  lengthInBytes)
{
  appConnQry* qry;
  appConnQryRF* rfDocId;

  rfDocId = (appConnQryRF*)safeMalloc(sizeof(appConnQryRF));
  rfDocId->docId = safeStrdup(docId);
  rfDocId->elemId = safeStrdup(elemId);
  rfDocId->varId = safeStrdup(varId);
  rfDocId->startByte = startByte;
  rfDocId->lengthInBytes = lengthInBytes;

  qry = (appConnQry*)safeMalloc(sizeof(appConnQry));
  qry->type = qryLeafRFDocId;
  qry->term = stringToAny(docId);
  qry->rf = rfDocId;

  return(qry);
}

/*---------------------------------------------------------------------------*/

appConnQry* 
appConnQryMakeNode(long operator,appConnQry* arg1,appConnQry* arg2)
/* these routines construct an appConnQrys, which are either a nodes or leaves
   of a query tree.  Leaves use qryLeaf values for their type, and operators 
   use qryNode types.  By calling these routines arbitrary query trees can 
   be built.

   Note makes a new appConnQry, but does not make copies of the arguments
 */
{
  appConnQry* qry;

  qry = (appConnQry*)safeMalloc(sizeof(appConnQry));
  qry->type = qrySubTree;
  qry->operator = operator;
  qry->arg1 = arg1;
  qry->arg2 = arg2;

  return(qry);
}

/*---------------------------------------------------------------------------*/

void 
appConnQryFree(appConnQry* qry)
/* this routine frees an appConnQry and any sub qry's or terms */
{
  if (qry == NULL)
    return;

  if (qry->type == qryLeafRFRecId)
   { appConnQryRF* rfRecId = qry->rf;
     freeAny(rfRecId->recId);
     safeFree(rfRecId->elemId);
     safeFree(rfRecId->varId);
     safeFree(rfRecId);
    }
  else if (qry->type == qryLeafRFDocId)
   { appConnQryRF* rfDocId = qry->rf;
     safeFree(rfDocId->docId);
     safeFree(rfDocId->elemId);
     safeFree(rfDocId->varId);
     safeFree(rfDocId);
    }    
  else if (qry->type == qrySubTree)
   { appConnQryFree(qry->arg1);
     appConnQryFree(qry->arg2);
   }

  freeAny(qry->term);
  safeFree(qry);
}

/*---------------------------------------------------------------------------*/
/* Status */
/*---------------------------------------------------------------------------*/

void
appConnStatusLogStatus(appConnStatus status)
{
  char* statusInfo;
  
  switch (status)
   { case appConnStatusUndefinedConnectionType:
       statusInfo = "Undefined Connection Type"; break;
     case appConnStatusOK:                     
       statusInfo = "Status Ok"; break;
     case appConnStatusUndefinedFunction:
       statusInfo = "Undefined Function"; break;
     case appConnStatusBadArguments:
       statusInfo = "Bad Arguments"; break;
     case appConnStatusNotFinished:
       statusInfo = "Not Finished"; break;
     default: statusInfo = "Unknown Status";
     }
  waislog(WLOG_HIGH,WLOG_ERROR,"status condition %ld: %s.",
	  status,statusInfo);
}

/*---------------------------------------------------------------------------*/
/* Diagnostics */
/*---------------------------------------------------------------------------*/

diagTableEntry diagTable[appConnDiagNumber] =
{{WLOG_HIGH,WLOG_ERROR,"error opening connection"},           /* ConnErrOpen */
 {WLOG_HIGH,WLOG_ERROR,"error reading from connection"},      /* ConnErrRead */
 {WLOG_HIGH,WLOG_ERROR,"error writing to connection"},       /* ConnErrWrite */
 {WLOG_LOW, WLOG_CLOSE,"closing connection"},                  /* ConnClosed */
 {WLOG_HIGH,WLOG_ERROR,"server error"},                       /* ServerError */
 {WLOG_HIGH,WLOG_ERROR,"protocol error"},                   /* ProtocolError */
 {WLOG_HIGH,WLOG_ERROR,"client error"},                       /* ClientError */
 {WLOG_HIGH,WLOG_ERROR,"access denied"},                     /* AccessDenied */
 {WLOG_HIGH,WLOG_ERROR,"unsupported search"},           /* UnsupportedSearch */
 {WLOG_HIGH,WLOG_ERROR,"bad packet"},                           /* BadPacket */
 {WLOG_HIGH,WLOG_ERROR,"no more headlines"},              /* NoMoreHeadlines */
 {WLOG_HIGH,WLOG_ERROR,"no database specified"},      /* NoDatabaseSpecified */
 {WLOG_HIGH,WLOG_ERROR,"no doc identifier specified"},   /* NoDocIdSpecified */
 {WLOG_HIGH,WLOG_ERROR,"no result set available"},   /* NoResultSetAvailable */
 {WLOG_HIGH,WLOG_ERROR,"no document found"},             /* DocumentNotFound */
 {WLOG_HIGH,WLOG_ERROR,"invalid chunk type"},            /* InvalidChunkType */
 {WLOG_HIGH,WLOG_ERROR,"invalid database name"},          /* InvalidDatabase */
 {WLOG_HIGH,WLOG_ERROR,"invalid doc identifier"},            /* InvalidDocId */
 {WLOG_LOW, WLOG_INFO, "request out of range"},         /* RequestOutOfRange */
 {WLOG_HIGH,WLOG_ERROR,"negotiated buffer size is too small"}
                                         /* appConnDiagNegotiatedBufTooSmall */
};

appConnDiag* 
appConnDiagMake(appConnDiagCode code, char* infoFormat,...)
/*  Note makes a new appConnDiag, and makes a copy of the infoFormat
    and its var_list.
 */
{
  appConnDiag* diag;
  char info[MAX_DIAG_INFO_LEN];
  va_list ap; 

  diag = (appConnDiag*)safeMalloc(sizeof(appConnDiag));
  diag->code = code;

  if (infoFormat != NULL)
   {
     va_start(ap,infoFormat);
     vsprintf(info,infoFormat,ap);
     va_end(ap); 
     diag->info = safeStrdup(info);
   }

  return(diag);
}

/*---------------------------------------------------------------------------*/

void
appConnDiagFree(appConnDiag* diag)
/* this routine frees an appConnDiag */
{
  if (diag == NULL)
    return;

  safeFree(diag->info);
  safeFree(diag);
}

/*---------------------------------------------------------------------------*/

void
appConnDiagLogErrors(list diags)
{ appConnDiag* diag;
  long i;

 for (i = 0, diag = nth(diags,0); 
	  i < length(diags); 
	  i++, diag = nth(diags,i))
  { diagTableEntry entry = diagTable[diag->code];
    waislog(entry.priority,entry.logType,"%s; %s",entry.message,diag->info);
  }

}

/*---------------------------------------------------------------------------*/
/* Variants */
/*---------------------------------------------------------------------------*/

appConnVariant* 
appConnMakeVariant(char* name,
		   char* composition,
		   char* language,
		   char* characterSet,
		   long  size,
		   long  cost,
		   long  currency)
/* here we make a variant structure, and copy in the arguments */
{
  appConnVariant* v;

  v = (appConnVariant*)safeMalloc(sizeof(appConnVariant));
  v->name = safeStrdup(name);
  v->composition = safeStrdup(composition);
  v->language = safeStrdup(language);
  v->characterSet = safeStrdup(characterSet);
  v->size = size;
  v->cost = cost;
  v->currency = currency;
  
  return(v);
}

/*---------------------------------------------------------------------------*/

void 
appConnVariantFree(appConnVariant* variant)
/* frees a variant and all it's sub fields */
{
  safeFree(variant->name);
  safeFree(variant->composition);
  safeFree(variant->language);
  safeFree(variant->characterSet);
  safeFree(variant);
}

/*---------------------------------------------------------------------------*/

void
appConnVariantListFree(list variants)
/* simply frees each variant in the list, and the list */
{
  long i;
  for (i = 0; i < length(variants); i++)
    appConnVariantFree((appConnVariant*)nth(variants,i));
  freeList(variants);
}

/*---------------------------------------------------------------------------*/
/* Elements */
/*---------------------------------------------------------------------------*/

appConnElement* 
appConnMakeElement(char* name,char* usageRight,list variants,list subElements)
/* this makes an element structure, and copies in the name and usageRight.
   it does not copy the variants or subElements
 */
{
  appConnElement* e;

  e = (appConnElement*)safeMalloc(sizeof(appConnElement));
  e->name = safeStrdup(name);
  e->usageRight = safeStrdup(usageRight);
  e->variants = variants;
  e->subElements = subElements;

  return(e);
}

/*---------------------------------------------------------------------------*/

void 
appConnElementFree(appConnElement* element)
/* deletes the element and all of it's sub structure */
{
  safeFree(element->name);
  safeFree(element->usageRight);
  appConnVariantListFree(element->variants);
  appConnElementListFree(element->subElements);
  safeFree(element);
}

/*---------------------------------------------------------------------------*/

void
appConnElementListFree(list elements)
/* simply frees each element in the list and the list itself */
{
  long i;
  for (i = 0; i < length(elements); i++)
    appConnElementFree((appConnElement*)nth(elements,i));
  freeList(elements);
}

/*---------------------------------------------------------------------------*/
/* Headlines */
/*---------------------------------------------------------------------------*/

appConnHeadline* 
appConnMakeHeadline(char* headline,
		    char* name,
		    date* date,
		    long  rank,
		    long  score,
		    any*  recId,
		    char* urx,
		    list  elements)
{
  appConnHeadline* hl;
  
  hl = (appConnHeadline*)safeMalloc(sizeof(appConnHeadline));
  hl->headline = safeStrdup(headline);
  hl->name = safeStrdup(name);
  hl->date = copyDate(date);
  hl->rank = rank;
  hl->score = score;
  hl->recId = duplicateAny(recId);
  hl->urx = safeStrdup(urx);
  hl->elements = elements;
  
  return(hl);
}

/*---------------------------------------------------------------------------*/

void 
appConnHeadlineFree(appConnHeadline* hl)
{
  safeFree(hl->headline);
  safeFree(hl->name);
  safeFree(hl->date);
  freeAny(hl->recId);
  safeFree(hl->urx);
  appConnElementListFree(hl->elements);

  safeFree(hl);
}

/*---------------------------------------------------------------------------*/
/* PDU dispatching */
/*---------------------------------------------------------------------------*/

appConnStatus 
appConnPduInfo(appConn* conn,
	       long*    pduType,
	       long*    packetLength,
	       list*    diags)
/* reads enough of the data on the connection to determine the pdu type 
   of the data on the line.  The data read is stored in the connection's
   buffer.  This routine setd the value of pduType, which is then used to 
   dispatch the pdu to the apropriate pdu reading function.  The data on the 
   line must be at the start of a pdu.

   internally, the routine may need to determine the pdu length.  this is
   done by:
   -reading wais header (v1)
   -reading z3950 header (v2)
   -reading till end of pdu (v2 indeterminite length)
*/
{
  appConnType* ai;

  if (conn == NULL || 
      (ai = findAppConnTypeNamed(conn->name)) == NULL ||
      ai->pduInfoFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
   { appConnStatus res;
     res = (*ai->pduInfoFunc)(conn,pduType,packetLength,diags);
     return(res);
   }
}

/*---------------------------------------------------------------------------*/
/* Utilities */
/*---------------------------------------------------------------------------*/

boolean
timeOut(time_t startTime,long timeOutSecs)
/* should this be in a subsecond resolution? */
{
  time_t timeNow = time(NULL);
  if (startTime + timeOutSecs < timeNow)
    return(true);
  else
    return(false);
}

/*---------------------------------------------------------------------------*/
