/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Module:  nConnMgr.c

 Purpose: This file is an interface to the various network abstractions.
          The idea is to make it easy to add new network abstractions.

          To add a new network type:
          1. write the creation functions.  Need to define one or both of
          the following:

          nConnStatus myNetConnOpenFunc(nConn** conn,long numArgs,va_list ap);
          nConnStatus myNetConnWaitForConnectionFunc(nConn** conn,
                                                     long numArgs,va_list ap);

          The functions malloc an structure which is based on an nConn (ie.
          it may be an nConn or it may be a bigger structure whose first 
          variable is an nConn) and assign it's pointer to *conn.  They then 
          assign read/write/free/etc functions to the fields in the nConn, 
          and interpret their arguments using the function va_arg() defined in 
          the ansi stdarg package.  These arguments are used to open or wait 
          for a connection to a remote system.  When the connection is 
          established, or an error occurs, nConnStatus is returned.

          2. Include the header file for the new connection in this file 
          just before the nConnTypeRegister() function.

          3. In nConnTypesInit(), add a call to nConnTypeRegister() for your
          new network abstraction.
----------------------------------------------------------------------------*/

#include "nConnMgr.h"
#include "cutil.h"
#include "list.h"

/*-------------------------------------------------------------------------- */

/* include header files for new connections here */
#include "tcpConn.h"
#include "stdioConn.h"

/*-------------------------------------------------------------------------- */

static list nConns = NULL;

/*-------------------------------------------------------------------------- */

static void
nConnTypeRegister(char* name,
		  nConnStatus (*openFunc)(nConn** conn,
					  long numArgs,
					  va_list ap),
		  nConnStatus (*waitForConnectionFunc)(nConn** conn,
						       long numArgs,
						       va_list ap),
		  nConnStatus (*initializeFunc)(nConn* c),
		  nConnStatus (*readFunc)(nConn* c,unsigned char* val),
		  nConnStatus (*peekFunc)(nConn* c,unsigned char* val),
		  nConnStatus (*writeFunc)(nConn* c,unsigned char val),
		  nConnStatus (*flushFunc)(nConn* c),
		  nConnStatus (*freeFunc)(nConn* c),
		  nConnStatus (*getIdFunc)(nConn* c,char** id,char** name))
{
  nConnType* ni;
  ni = (nConnType*)safeMalloc(sizeof(nConnType));
  ni->name = safeStrdup(name);
  ni->openFunc = openFunc;
  ni->waitForConnectionFunc = waitForConnectionFunc;
  ni->initializeFunc = initializeFunc;
  ni->readFunc = readFunc;
  ni->peekFunc = peekFunc;
  ni->writeFunc = writeFunc;
  ni->flushFunc = flushFunc;
  ni->freeFunc = freeFunc;
  ni->getIdFunc = getIdFunc;
  nConns = collecting(nConns,ni);
}

/*---------------------------------------------------------------------------*/

void 
nConnTypesInit(void)
{
  if (nConns == NULL)
   { /* register connections here */
     nConnTypeRegister("TCP",
		       tcpConnOpen,
		       tcpConnWaitForConnection,
		       tcpConnInitialize,
		       tcpConnRead,
		       tcpConnPeek,
		       tcpConnWrite,
		       tcpConnFlush,
		       tcpConnFree,
		       tcpConnGetId);
     nConnTypeRegister("stdio",
		       stdioConnOpen,
		       stdioConnWaitForConnection,
		       stdioConnInitialize,
		       stdioConnRead,
		       stdioConnPeek,
		       stdioConnWrite,
		       stdioConnFlush,
		       stdioConnFree,
		       stdioConnGetId);
   }
}

/*-------------------------------------------------------------------------- */

void
nConnTypesFree(void)
{
  nConnType* ni;
  long i;

  for(i=0; i<length(nConns); i++)
    {
       ni = nth(nConns,i);
       safeFree(ni->name);
       safeFree(ni);
     }

  freeList(nConns);
  nConns = NULL;
}

/*-------------------------------------------------------------------------- */

nConnType* 
findNConnTypeNamed(char* name)
{
  long i;

  for (i = 0; i < length(nConns); i++)
   { nConnType* ni = (nConnType*)nth(nConns,i);
     if (substrcmp(ni->name,name))
       return(ni);
   }
  return(NULL);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnOpen(nConn** conn,
	  char* connType,
	  long numArgs,
	  ...)
{
  va_list ap;	
  nConnStatus res;
  nConnType* ni = findNConnTypeNamed(connType);

  if (ni == NULL || ni->openFunc == NULL || conn == NULL)
    return(nConnStatusUndefinedConnectionType);

  va_start(ap,numArgs);	
  res = (*ni->openFunc)(conn,numArgs,ap);
  if (*conn != NULL)
    (*conn)->name = safeStrdup(connType);
  va_end(ap);	     

  if (res == nConnStatusOK)
    res = nConnInitialize(*conn);

  return(res);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnWaitForConnection(nConn** conn,
		       char* connType,
		       long numArgs,
		       ...)
{
  va_list ap;	
  nConnStatus res;
  nConnType* ni = findNConnTypeNamed(connType);

  if (ni == NULL || ni->waitForConnectionFunc == NULL || conn == NULL)
    return(nConnStatusUndefinedConnectionType);

  va_start(ap,numArgs);	
  res = (*ni->waitForConnectionFunc)(conn,numArgs,ap);
  if (*conn != NULL)
    (*conn)->name = safeStrdup(connType);
  va_end(ap);	     

  if (res == nConnStatusOK)
    res = nConnInitialize(*conn);

  return(res);
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnInitialize(nConn* conn)
{
  nConnType* netType;

  if (conn == NULL)
    return(nConnStatusBadArguments);
  else if ((netType = findNConnTypeNamed(conn->name)) == NULL)
    return(nConnStatusUndefinedConnectionType);
  else if (netType->initializeFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
    return((*netType->initializeFunc)(conn));
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnRead(nConn* conn,
	  unsigned char* val)
{
  nConnType* netType;

  if (conn == NULL)
    return(nConnStatusBadArguments);
  else if ((netType = findNConnTypeNamed(conn->name)) == NULL)
    return(nConnStatusUndefinedConnectionType);
  else if (netType->readFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
    return((*netType->readFunc)(conn,val));
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnPeek(nConn* conn,
	  unsigned char* val)
{
  nConnType* netType;

  if (conn == NULL)
    return(nConnStatusBadArguments);
  else if ((netType = findNConnTypeNamed(conn->name)) == NULL)
    return(nConnStatusUndefinedConnectionType);
  else if (netType->peekFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
    return((*netType->peekFunc)(conn,val));
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnWrite(nConn* conn,
	   unsigned char val)
{
  nConnType* netType;

  if (conn == NULL)
    return(nConnStatusBadArguments);
  else if ((netType = findNConnTypeNamed(conn->name)) == NULL)
    return(nConnStatusUndefinedConnectionType);
  else if (netType->writeFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
    return((*netType->writeFunc)(conn,val));
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnFree(nConn* conn)
{
  nConnType* netType;

  if (conn == NULL)
    return(nConnStatusBadArguments);
  else if ((netType = findNConnTypeNamed(conn->name)) == NULL)
    return(nConnStatusUndefinedConnectionType);
  else if (netType->freeFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
    return((*netType->freeFunc)(conn));
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnFlush(nConn* conn)
{
  nConnType* netType;

  if (conn == NULL)
    return(nConnStatusBadArguments);
  else if ((netType = findNConnTypeNamed(conn->name)) == NULL)
    return(nConnStatusUndefinedConnectionType);
  else if (netType->flushFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
    return((*netType->flushFunc)(conn));
}

/*-------------------------------------------------------------------------- */

nConnStatus 
nConnGetId(nConn* conn,
	   char** netId,
	   char** name)
{
  nConnType* netType;

  if (conn == NULL)
    return(nConnStatusBadArguments);
  else if ((netType = findNConnTypeNamed(conn->name)) == NULL)
    return(nConnStatusUndefinedConnectionType);
  else if (netType->getIdFunc == NULL)
    return(nConnStatusUndefinedFunction);
  else
    return((*netType->getIdFunc)(conn,netId,name));
}

/*-------------------------------------------------------------------------- */
