/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Module:   nConnMgr.h

 Purpose:  This file defines an abstract way of creating and accessing a
           remote connection, where the connection is independent of the 
           underlying communications protocol.  The remote connection could,
           for example, be TCP/IP, stdio, IPX, etc.

 Note:     In this release, the network connections "TCP" and "stdio" are
           the registered types of network connections.
----------------------------------------------------------------------------*/

#ifndef nConnMgr_h
#define nConnMgr_h

#include "cutil.h"



/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 DataType:  nConnStatus
 Purpose:   The nConnStatus data type is used for status returned from the
            Network Connection API functions.  The corresponding constants
            are the possible values that the nConnStatus data type may assume.
----------------------------------------------------------------------------*/
typedef long nConnStatus;

#define nConnStatusUndefinedConnectionType (-1L)
#define nConnStatusOK                       (0L)
#define nConnStatusUndefinedFunction        (1L)
#define nConnStatusBadArguments             (2L)
#define nConnStatusOpenFailed               (3L)
#define nConnStatusWriteFailed              (4L)
#define nConnStatusReadFailed               (5L)
#define nConnStatusReadEOF                  (6L)
#define nConnStatusReadNoData               (7L)
#define nConnStatusNoIDAvailable            (8L)

/*----------------------------------------------------------------------------
 Datatype: nConn
 Purpose:  The nConn data type is the default data structure used to contain
           information on a specific instance of a connection.  It contains
           the state of the local connection.
 ----------------------------------------------------------------------------*/
typedef struct nConn
{ char* name;
  boolean lastOpWasRead;
} nConn;




/*----------------------------------------------------------------------------
  ----------------------------------------------------------------------------
 Datatype: nConnType
 Purpose:  The nConnType data type contains an extensible table of function
           pointers in the API.  These include a connection open and a wait
           for connection API function.  There is one instance of this table
           for each type of protocol supported.
 ----------------------------------------------------------------------------*/
typedef struct nConnType
{ char* name;
  nConnStatus(*openFunc)(nConn** conn,long numArgs,va_list ap);
  nConnStatus(*waitForConnectionFunc)(nConn** conn,long numArgs,va_list ap);
  nConnStatus(*initializeFunc)(nConn* c);
  nConnStatus(*readFunc)(nConn* c,unsigned char* val);
  nConnStatus(*peekFunc)(nConn* c,unsigned char* val);
  nConnStatus(*writeFunc)(nConn* conn,unsigned char val);
  nConnStatus(*flushFunc)(nConn* c);
  nConnStatus(*freeFunc)(nConn* c);
  nConnStatus(*getIdFunc)(nConn* c,char** id,char** name);
} nConnType;

/*----------------------------------------------------------------------------
 Function:    nConnTypesInit
 Purpose:     To register all types of network connections.
 Called by:   Called by client or server program.
 Globals:     The registered types are collected and maintained in a global
              list, nConns.
 Parameters:  none
 Returns:     void
----------------------------------------------------------------------------*/
void nConnTypesInit(void);

/*----------------------------------------------------------------------------
 Function:    nConnTypesFree
 Purpose:     To free the registered network connections.
 Called by:   Called by client or server program.
 Globals:     The registered types maintained in a global list, nConns, are
              freed.
 Parameters:  none
 Returns:     void
----------------------------------------------------------------------------*/
void nConnTypesFree(void);


nConnStatus nConnOpen(nConn** conn,char* connType,long numArgs,...);
nConnStatus nConnWaitForConnection(nConn** conn,char* connType,
				   long numArgs,...);
nConnStatus nConnInitialize(nConn* c);
nConnStatus nConnRead(nConn* c,unsigned char* val);
nConnStatus nConnPeek(nConn* c,unsigned char* val);
nConnStatus nConnWrite(nConn* c,unsigned char val);
nConnStatus nConnFlush(nConn* c);
nConnStatus nConnFree(nConn* c);
nConnStatus nConnGetID(nConn* c,char** id,char** name);



#endif /* ndef nConnMgr_h */

