#define sockets_c

#include "sockets.h"

#include <errno.h>
#include <string.h>
#include <arpa/inet.h>
#include <signal.h>

#include "panic.h"
#include "waislog.h"
#include "list.h"

extern int errno;

extern char *sys_errlist[];

#define QUEUE_SIZE (3L)
#define HOSTNAME_BUFFER_SIZE (120L)
#define MAX_RETRYS (10L)

/*---------------------------------------------------------------------------*/

static boolean 
clrSocket(struct sockaddr_in *address,long portnumber,long *sock)
{
  if (errno == EADDRINUSE) 
   { if (connect(*sock, address, sizeof (struct sockaddr_in)) == 0) 
      { close(*sock);
	waislog(WLOG_HIGH,WLOG_ERROR,
		"cannot bind port %ld (address already in use)",
		portnumber);
	waislog(WLOG_HIGH,WLOG_ERROR,
		"waisserver is already running on this system");
	panic("Exiting");
      } 
     else 
      { int one = 1;
	
	close(*sock);
	if ((*sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	  panic("Open socket failed in trying to clear the port.");
	if (setsockopt(*sock,SOL_SOCKET,SO_REUSEADDR,&one,sizeof(one)) < 0) 
	  ; /* do nothing? */
	address->sin_family = AF_INET;
	address->sin_addr.s_addr = INADDR_ANY;
	address->sin_port = htons(portnumber);
	if (bind(*sock, address, sizeof(*address)) == 0) 
	 ; /* do nothing? */
      }
   }
  return(true);
}

/*---------------------------------------------------------------------------*/

void
openServer(long port,long* fd,long size)
{ 
  struct sockaddr_in address;

  memset(&address, 0, sizeof(address));

  if ((*fd = socket(AF_INET,SOCK_STREAM,0)) < 0)
   { panic("can't get file descriptor for socket: %s", sys_errlist[errno]);
   }

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = htonl(INADDR_ANY);
  address.sin_port = htons(port);

  if (bind(*fd,(struct sockaddr*)&address,sizeof(struct sockaddr)) < 0)
    clrSocket(&address, port, fd);

  if (listen(*fd,QUEUE_SIZE) < 0)
    panic("can't open server: %s", sys_errlist[errno]);
}

/*---------------------------------------------------------------------------*/

void
fdAcceptClientConnection(long socket,long* fd)
{ 
  struct sockaddr_in source;
  int sourcelen;

#ifdef BSD
  struct in_addr {
    union {
      struct { u_char s_b1,s_b2,s_b3,s_b4; } S_un_b;
      u_long S_addr;
    } S_un;
  } addr_p;
#endif 

  sourcelen = sizeof(struct sockaddr_in);

  do {
       errno = 0;
       *fd = accept(socket, &source, &sourcelen);
     } while (*fd < 0 && errno == EINTR);

  if (source.sin_family == AF_INET) 
   { struct hostent *peer = NULL;

     peer = gethostbyaddr((char*)&source.sin_addr, 4, AF_INET);

#ifdef obsolete
     waislog(WLOG_MEDIUM,WLOG_CONNECT,"accepted connection from: %s [%s]",
	     (peer == NULL) ? "" : peer->h_name,
#if defined(sparc) && defined(__GNUC__) && !defined(__svr4__)
	     inet_ntoa(&source.sin_addr));
#else
             inet_ntoa(source.sin_addr));
#endif /* sparc */
#endif /* def obsolete */
   }

  if (*fd < 0)
    panic("can't accept connection");
}

/*---------------------------------------------------------------------------*/

void
acceptClientConnection(long socket,FILE** file)
{ 
  long fd; 
  fdAcceptClientConnection(socket,&fd);
  if ((*file = fdopen(fd,"r+")) == NULL)
    panic("can't accept connection");
}

/*---------------------------------------------------------------------------*/

void
closeClientConnection(FILE* file)
{ 
  fclose(file);
}

/*---------------------------------------------------------------------------*/

void
closeServer(long socket)
{
  close(socket);
}

/*---------------------------------------------------------------------------*/

boolean
fdConnectToServer(char* host_name,long port,long* fd)
{
  char hostnamebuf[80];
  long rc, i;
  struct hostent *host;
  
  struct sockaddr_in name;

  memset((char *)&name, 0,sizeof (name));

#ifdef DG
    { struct in_addr tmp = inet_addr(host_name);
      name.sin_addr.s_addr = tmp.s_addr;
    }
#else
  name.sin_addr.s_addr = inet_addr(host_name);
#endif

  if (name.sin_addr.s_addr != -1) 
   { name.sin_family = AF_INET;
     (void) strcpy(hostnamebuf, host_name);
   }
  else 
   { host = gethostbyname(host_name);

     if(NULL == host)
      return(false);

     name.sin_family = host->h_addrtype;
#ifdef h_addr
     memcpy((caddr_t)&name.sin_addr,host->h_addr_list[0],host->h_length);
#endif
     (void) strcpy(hostnamebuf, host->h_name);
   }
  host_name = hostnamebuf;

  name.sin_port = htons(port);

  *fd = socket (AF_INET, SOCK_STREAM, 0);
  for (i = 0; i < MAX_RETRYS; i++) 
   { int mask = sigmask(SIGALRM);

     mask = sigblock(mask);
     rc = connect(*fd, &name, sizeof (name));
     sigsetmask(mask);

    if (rc == 0) 
      return(true);
    else if (errno == EINTR)
      sleep(1);
    else break;
  }

  perror("Connect to socket did not work");
  return(false);
}

/*---------------------------------------------------------------------------*/

FILE *
connectToServer(char* host_name,long port)
{
  FILE* file;
  long fd;
  if (fdConnectToServer(host_name,port,&fd) == false) 
    return(NULL);

  if ((file = fdopen(fd,"r+")) == NULL) 
   { perror("Connect to socket did not work, fdopen failure");
     return(NULL);
   }

  return(file);
}

/*---------------------------------------------------------------------------*/

void
closeConnectionToServer(FILE* file)
{
  fclose(file);
}

/*---------------------------------------------------------------------------*/

boolean
isSameIPAddress(char* host1,char* host2)
/* return true if the machines are the same, as determined by their ip 
   address.  False otherwise
 */
{
  struct hostent* addr1;
  struct hostent* addr2;
  char* ip1;
  char* ip2;
  long i1,i2;
  list l1 = NULL;
  boolean res = false;

  if (host1 == NULL || host2 == NULL)
    return(false);

  /* get the ip addresses of the first host */
  addr1 = gethostbyname(host1);
  /* copy it, since another call to gethostbyname will stomp the results */
  for (i1 = 0; addr1 != NULL; i1++)
   { ip1 = addr1->h_addr_list[i1];
     if (ip1 == NULL)
       break; /* done copying */
     l1 = collecting(l1,safeStrdup(ip1));
   }
  
  /* get the second address */
  addr2 = gethostbyname(host2);

  /* compare them, go down the addresses in addr1, and see if any of them
     match addresses in addr2
   */
  for (i1 = 0; addr2 != NULL && i1 < length(l1); i1++)
   { ip1 = (char*)nth(l1,i1);
     for (i2 = 0; true; i2++)
      { ip2 = addr1->h_addr_list[i2];
	if (ip2 == NULL)
	  break; /* no more addresses in addr2 try next entry in addr1 */
	if (strcmp(ip1,ip2) == 0)
	 { res = true; /* it's the same machine! */
	   break;
	 }
      }
   }

  for (i1 = 0; i1 < length(l1); i1++)
   { ip1 = (char*)nth(l1,i1);
     safeFree(ip1);
   }
  freeList(l1);

  return(res);
}

/*---------------------------------------------------------------------------*/
