/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Module:   cli.c

 Purpose:  This modules provides a generic layer of function calls to the
           z39.50 V2 protocol.  It provides the basic initialization, search,
           and present services of z39.50.

           Use the DEBUG switch to dump the outgoing and incoming 
           packets to a file.  By default, this is set to OFF.
---------------------------------------------------------------------------*/

#include <string.h>
#include "cutil.h"
#include "list.h"
#include "appConnCli.h"

#include "connV2Cli.h"
#include "berutil.h"
#include "tag.h" 
#include "cli.h"
#include "rpnBld.h"
#include "recPrs.h"

/* When DEBUG is defined, all APDU's (application protocol data units) are
   written to a file "LogFile" in the current working directory.
#define DEBUG
*/



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2InitReqWriteAPDU
 Purpose:    Write an initialize request PDU using BER to the buffer pBuf 
             with length pBufLen, and check for errors.
 Called by:  Z3950V2InitReqWrite
 Parameters: pBuf        -- output;
             pBufLen     -- output;
             pPrefMsgSiz -- input;
             pMaxRecSiz  -- input;
             pIdAuth     -- input;
             pImpId      -- input;
             pImpName    -- input;
             pImpVer     -- input;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
 Imp Notes:  The following Z39.50 features are not supported in this release
             of the initialization request message:
             1. Authentication is ignored.
             2. User information field is ignored.  In the future, the
             variable arguments will be used to hold this information.  
             Anticipated uses are for character set negotiation records.
---------------------------------------------------------------------------*/
boolean
Z3950V2InitReqWriteAPDU(char* pBuf,
			long* pBufLen,
			long  pPrefMsgSiz, 
			long  pMaxRecSiz, 
			char* pIdAuth,
			char* pImpId, 
			char* pImpName, 
			char* pImpVer, 
			...)
{ 
  DATA_DIR *lTop = NULL;	/* Top-level directory structure. */
  boolean lError = false;

  /* Create a directory structure. */
  if ((lTop = dmake(tagPDUInitReq,ASN1_CONTEXT,20L)) == NULL)
    lError = true;

  /* Required protocol version: Z39.50 Versions 1 & 2. */
  /* Required options: Search and Present Services. */
  if (lError ||
      (daddbits(lTop,tagInitProtVer,ASN1_CONTEXT,Z3950V2_PROT_VER) == NULL) ||
      (daddbits(lTop,tagInitOpts,ASN1_CONTEXT,Z3950V2_OPTIONS) == NULL))
    lError = true;

  /* Optional authentication information. */
  /* XXX Authentication is not handled in this release. */

  /* Required preferred message and maximum record size. */
  if (lError ||
      (daddnum(lTop,tagInitPrefMsgSiz,ASN1_CONTEXT, 
	       (CHAR*)&pPrefMsgSiz,sizeof(pPrefMsgSiz)) == NULL) ||
      (daddnum(lTop,tagInitMaxRecSiz,ASN1_CONTEXT, 
	       (CHAR*)&pMaxRecSiz,sizeof(pMaxRecSiz)) == NULL))
    lError = true;

  /* Optional implementation identification, name, version, and userInfo. */
  if (lError ||
      ((pImpId != NULL) &&
       ((daddchar(lTop,tagInitImpId,ASN1_CONTEXT, 
		  (CHAR*)pImpId,safeStrlen(pImpId))) == NULL)) ||
      ((pImpName != NULL) &&
       ((daddchar(lTop,tagInitImpName,ASN1_CONTEXT, 
		  (CHAR*)pImpName,safeStrlen(pImpName))) == NULL)) ||
      ((pImpVer != NULL) &&
       ((daddchar(lTop,tagInitImpVer,ASN1_CONTEXT, 
		  (CHAR*)pImpVer,safeStrlen(pImpVer))) == NULL)))
    lError = true;

#ifdef DEBUG
  if (!lError)
    BufDumpBer("Client InitRequest",lTop);
#endif

  if (!lError)
   {
     /* Determine length, and build pBuf from directory. */
     rec_len(lTop);
     *pBufLen = asm_rec(lTop,(CHAR*)pBuf);
   }
  else *pBufLen = 0L;

  if (lTop != NULL) dfree(lTop);
  return(!lError);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2InitRespReadAPDU
 Purpose:    Read an initialize response PDU in BER from the buffer pBuf, and
             check for errors.
             For this release, the user information field is ignored.  In the
             future, the variable arguments will be used to hold this info.
             Anticipated uses are for character set negotiation records.
 Called by:  Z3950V2InitRespRead
 Parameters: pBuf        -- input;
             pPrefMsgSiz -- output;
             pRes        -- output;
             pImpId      -- output;
             pImpName    -- output;
             pImpVer     -- output;
             pDiags      -- output;
 Returns:    true(1) or false(0) depending on whether reading the PDU was 
             successful.
---------------------------------------------------------------------------*/
boolean
Z3950V2InitRespReadAPDU(char*    pBuf, 
			long*    pPrefMsgSiz,
			long*    pMaxRecSiz, 
			boolean* pRes, 
			char**   pImpId,
			char**   pImpName, 
			char**   pImpVer, 
			list*    pDiags,
			...)
{ 
  DATA_DIR *lTop = NULL;
  DATA_DIR *lNxt = NULL;
  DATA_DIR *lTmp = NULL;
  boolean lError = false;

  /* Read the buffer into lTop, and check if successful. */
  if (((lTop = dalloc(20)) == NULL) ||
      (bld_dir((CHAR*)pBuf, lTop) == false) ||
      ((lNxt = tagFindNext(lTop,tagPDUInitResp)) == NULL))
    lError = true;

#ifdef DEBUG     
  if (!lError)
    BufDumpBer("Client InitResponse",lTop);
#endif

  /* Begin decoding the init response sequence. */

  /* XXX Need to check the version and the options. */

  /* If any of the required arguments are NULL (unspecified)
     the pdu is non-conformant with the standard. */
  /* Read the required preferred message size. */
  if (!lError && ((lNxt = tagFindChild(lNxt,tagInitPrefMsgSiz)) != NULL))
    *pPrefMsgSiz = dgetnum(lNxt);
  else lError = true;

  /* Read the required maximum record size. */
  if (!lError && ((lNxt = tagFindNext(lNxt,tagInitMaxRecSiz)) != NULL))
    *pMaxRecSiz = dgetnum(lNxt);
  else lError = true;

  /* Read the required result. */
  if (!lError && ((lNxt = tagFindNext(lNxt,tagInitRes)) != NULL))
    *pRes = dgetnum(lNxt);
  else lError = true;

  /* Read optional implementation identification. */
  if (!lError && 
      (pImpId != NULL) &&
      ((lTmp = tagFindNext(lNxt,tagInitImpId)) != NULL))
    *pImpId = safeStrndup((char*)lTmp->ptr.data,lTmp->count+1);

  /* Read optional implementation name. */
  if (!lError &&
      (pImpName != NULL) &&
      ((lTmp = tagFindNext(lNxt,tagInitImpName)) != NULL))
    *pImpName = safeStrndup((char*)lTmp->ptr.data,lTmp->count+1);

  /* Read optional implementation version. */
  if (!lError &&
      (pImpVer != NULL) &&
      ((lTmp = tagFindNext(lNxt,tagInitImpVer)) != NULL))
    *pImpVer = safeStrndup((char*)lTmp->ptr.data,lTmp->count+1);

  if (lTop != NULL) dfree(lTop);
  return(*pRes = !lError);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2SearchReqWriteAPDU
 Purpose:    Write a search request PDU using BER to the buffer pBuf with
             length pBufLen, and check for errors.
 Called by:  Z3950V2SearchReqWrite
 Parameters: pBuf               -- output;
             pBufLen            -- output;
             pSmlSetUpBnd       -- input;
             pLrgSetLoBnd       -- input;
             pReplInd           -- input;
             pResSetName        -- input;
             pDbName            -- input;
             pSmlSetElemSetName -- input;
             pMedSetElemSetName -- input;
             pPrefRecSyn        -- input;
             pQry               -- input;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
Z3950V2SearchReqWriteAPDU(char*       pBuf,
			  long*       pBufLen,
			  long        pSmlSetUpBnd,
			  long        pLrgSetLoBnd, 
			  long        pMedSetPresNum,
			  boolean     pReplInd,
			  char*       pResSetName,
			  char*       pDbName,
			  char*       pSmlSetElemSetName, 
			  char*       pMedSetElemSetName,
			  char*       pPrefRecSyn, 
			  appConnQry* pQry)
{ 
  DATA_DIR *lTop = NULL;
  DATA_DIR *lElm = NULL;
  DATA_DIR *lGen = NULL;
  boolean lError = false;
  
  /* Create a directory structure. */
  if ((lTop = dmake(tagPDUSrchReq,ASN1_CONTEXT,20L)) == NULL)
    lError = true;

  /* Required small set upper and lower bounds, replace indicator, 
     and result set name. */
  if (lError ||
      (daddnum(lTop,tagSrchSmlSetUpBnd,ASN1_CONTEXT, 
	       (CHAR*)&pSmlSetUpBnd,sizeof(pSmlSetUpBnd)) == NULL) ||
      (daddnum(lTop,tagSrchLrgSetLoBnd,ASN1_CONTEXT, 
	       (CHAR*)&pLrgSetLoBnd,sizeof(pLrgSetLoBnd)) == NULL) ||
      (daddnum(lTop,tagSrchMedSetPrsntNum,ASN1_CONTEXT, 
	       (CHAR*)&pMedSetPresNum,sizeof(pMedSetPresNum)) == NULL) ||
      (daddnum(lTop,tagSrchReplInd,ASN1_CONTEXT, 
	       (CHAR*)&pReplInd,sizeof(pReplInd)) == NULL) ||
      (daddchar(lTop,tagSrchResSetName,ASN1_CONTEXT, 
		(CHAR*)pResSetName,safeStrlen(pResSetName)) == NULL))
    lError = true;

  /* Required database name. */
  if (lError ||
      ((lElm = daddtag(lTop,tagSrchDbNames,ASN1_CONTEXT)) == NULL) ||
      ((daddchar(lElm,tagGlobDbName,ASN1_CONTEXT,
		 (CHAR*)pDbName,safeStrlen(pDbName))) == NULL))
    lError = true;

  /* Optional small set element set names. */
  if (lError ||
      ((pSmlSetElemSetName != NULL) &&
       (((lElm = daddtag(lTop,tagSrchSmlSetElmtSetName,
			 ASN1_CONTEXT)) == NULL) ||
	((lGen= daddtag(lElm,tagElmtSetNamesGen,ASN1_CONTEXT)) == NULL) ||
	(daddchar(lGen,tagElmtSetName,ASN1_CONTEXT, 
		  (CHAR*)pSmlSetElemSetName,
		  safeStrlen(pSmlSetElemSetName))) == NULL)))
    lError = true;

  /* Optional medium set element set names. */
  if (lError ||
      ((pMedSetElemSetName != NULL) &&
       (((lElm = daddtag(lTop,tagSrchMedSetElmtSetName,
			 ASN1_CONTEXT)) == NULL) ||
	((lGen = daddtag(lElm,tagElmtSetNamesGen,ASN1_CONTEXT)) == NULL) ||
	(daddchar(lGen,tagElmtSetName,ASN1_CONTEXT, 
		  (CHAR*)pMedSetElemSetName,
		  safeStrlen(pMedSetElemSetName))) == NULL)))
    lError = true;

  /* Optional preferred record syntax. */
  if (lError ||
      ((pPrefRecSyn != NULL) &&
       (daddoid(lTop,tagSrchAndPrsntPrefRecSyn,
		ASN1_CONTEXT,pPrefRecSyn)) == NULL))
    lError = true;

  /* Build the Type-1 Query. */
  if (lError ||
      (rpnBld(lTop,pQry) == false))
    lError = true;

#ifdef DEBUG
  if (!lError)
    BufDumpBer("Client SearchRequest",lTop);
#endif

  if (!lError)
   {
     /* Determine length, and build pBuf from directory. */
     rec_len(lTop);
     *pBufLen = asm_rec(lTop,(CHAR*)pBuf);
   }
  else *pBufLen = 0L;

  if (lTop != NULL) dfree(lTop);
  return(!lError);
}
 


/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2SearchRespReadAPDU
 Purpose:    Read a search response PDU in BER from the buffer pBuf, and check
             for errors.  This function does not read the returned records or
             the diagnostics.
 Called by:  Z3950V2SearchRespReadVariantsAPDU
             Z3950V2SearchRespReadObjElemAPDU
 Parameters: pBuf          -- input;
             pResCnt       -- output;
             pNumRecRet    -- output;
             pNxtResSetPos -- output;
             pSrchStat     -- output;
             pResSetStat   -- output;
             pPrsntStat    -- output;
             pTop          -- output;
             pDiags        -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
Z3950V2SearchRespReadAPDU(char*      pBuf,
			  long*      pResCnt, 
			  long*      pNumRecRet,
			  long*      pNxtResSetPos, 
			  boolean*   pSrchStat,
			  long*      pResSetStat,
			  long*      pPrsntStat,
			  DATA_DIR** pTop,
			  list*      pDiags)
{
  DATA_DIR *lNxt = NULL;
  DATA_DIR *lStat = NULL;
  boolean lError = false;

  /* Check for errors in the search response PDU. */
  if (((*pTop = dalloc(20)) == NULL) ||
      (bld_dir((CHAR*)pBuf, *pTop) == false) ||
      ((lNxt = tagFindNext(*pTop,tagPDUSrchResp)) == NULL))
    lError = true;

#ifdef DEBUG
  if (!lError)
    BufDumpBer("Client SearchResponse",*pTop);
#endif DEBUG

  /* Begin decoding the search response sequence. */
  /* If any of the required arguments are NULL (unspecified)
     the pdu is non-conformant with the standard. */

  /* Read the result count. */
  if (!lError && ((lNxt = tagFindChild(lNxt,tagSrchResCnt)) != NULL))
    *pResCnt = dgetnum(lNxt);
  else lError = true;

  /* Read the number of records returned. */
  if (!lError && ((lNxt = tagFindNext(lNxt,tagSrchAndPrsntNumRecRet)) != NULL))
    *pNumRecRet = dgetnum(lNxt);
  else lError = true;

  /* Read the next result set position. */
  if (!lError && 
      ((lNxt = tagFindNext(lNxt,tagSrchAndPrsntNxtResSetPos)) != NULL))
    *pNxtResSetPos = dgetnum(lNxt);
  else lError = true;

  /* Read the search status. */
  if (!lError && ((lNxt = tagFindNext(lNxt,tagSrchSrchStat)) != NULL))
    *pSrchStat = dgetnum(lNxt);
  else lError = true;

  if (!lError)
   {
     /* Read the optional status parameters. */
     if (*pSrchStat)
      { 
	/* If the search status is true, read the present status. */
	if ((lStat = tagFindNext(lNxt,tagSrchAndPrsntPrsntStat)))
	  *pResSetStat = dgetnum(lStat);
      }
     else
      { 
	/* Otherwise, read the result set status. */
	if ((lStat = tagFindNext(lNxt,tagSrchResSetStat)))
	  *pResSetStat = dgetnum(lStat);
      }
   }
  else
   {
     /* Handle the error. */  
     /* Ensure search status is false, result set status is none, result count
	and number of records returned is zero. */
     *pSrchStat = false;
     *pResSetStat = tagResSetStatNon;
     *pResCnt = *pNumRecRet = 0L;
   }

  return(!lError);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2SearchRespReadVariantsAPDU
 Purpose:    Read a search response PDU in BER from the buffer pBuff, and
             check for errors.  Reads the full PDU.  Expects to read records
             from the "VARIANT" element set.
 Called by:  Z3950V2SearchRespReadVariants
 Parameters: pBuf          -- input;
             pResCnt       -- output;
             pNumRecRet    -- output;
             pNxtResSetPos -- output;
             pSrchStat     -- output;
             pResSetStat   -- output;
             pPresntStat   -- output;
             pRecs         -- output;
             pDiags        -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
Z3950V2SearchRespReadVariantsAPDU(char*    pBuf,
				  long*    pResCnt, 
				  long*    pNumRecRet,
				  long*    pNxtResSetPos, 
				  boolean* pSrchStat,
				  long*    pResSetStat,
				  long*    pPrsntStat,
				  list*    pRecs,
				  list*    pDiags)
{ 
  DATA_DIR *lTop = NULL;
  boolean lRetCode;

  lRetCode = 
    /* Read the buffer into lTop and decode. */
    (Z3950V2SearchRespReadAPDU(pBuf,pResCnt,pNumRecRet,pNxtResSetPos, 
			       pSrchStat,pResSetStat,pPrsntStat,
			       &lTop,pDiags) && 
     /* Decode the database or diagnostic records. */
     recPrsRecs(lTop,pDiags,VARIANT_ELEM_SET,pRecs));

  if (lTop != NULL) dfree(lTop);
  return(lRetCode);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2SearchRespReadObjElemAPDU
 Purpose:    Read a search response PDU in BER from the buffer pBuff, and
             check for errors.  Reads the full PDU.  Expects to read the full
             object element from the element set; this may be in the form of
             a full text, image, or multi-media object.
 Called by:  Z3950V2SearchRespReadObjElem
 Parameters: pBuf          -- input;
             pResCnt       -- output;
             pNumRecRet    -- output;
             pNxtResSetPos -- output;
             pSrchStat     -- output;
             pResSetStat   -- output;
             pPrsntStat    -- output;
             pElem         -- output;
             pObjElem      -- output;
             pObjElemLen   -- output;
             pTargetPtr    -- output;
             pDiags        -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
Z3950V2SearchRespReadObjElemAPDU(char*    pBuf,
				 long*    pResCnt, 
				 long*    pNumRecRet,
				 long*    pNxtResSetPos, 
				 boolean* pSrchStat,
				 long*    pResSetStat,
				 long*    pPrsntStat,
				 char**   pElem,
				 char**   pObjElem, 
				 long*    pObjElemLen,
				 char**   pTargetPtr,
				 list*    pDiags)
{ 
  DATA_DIR *lTop = NULL;
  boolean lRetCode = true;

  lRetCode =
    /* Read the buffer into lTop and decode. */
    (Z3950V2SearchRespReadAPDU(pBuf,pResCnt,pNumRecRet,pNxtResSetPos, 
			       pSrchStat,pResSetStat,pPrsntStat,
			       &lTop,pDiags) &&
     /* Decode the database or diagnostic records. */
     recPrsRecs(lTop,pDiags,OBJ_ELEM_ELEM_SET,
		pElem,pObjElem,pObjElemLen,pTargetPtr));

  if (lTop != NULL) dfree(lTop);
  return(lRetCode);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2PresentReqWriteAPDU
 Purpose:    Write a present request PDU using BER to the buffer pBuf with
             length pBufLen, and check for errors.
 Called by:  Z3950V2PresentReqWrite
 Parameters: pBuf          -- output;
             pBufLen       -- output;
             pResSetId     -- input;
             pResSetStrtPt -- input;
             pNumRecReq    -- input;
             pElemSetName  -- input;
             pPrefRecSyn   -- input;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
Z3950V2PresentReqWriteAPDU(char* pBuf,
			   long* pBufLen, 
			   char* pResSetId, 
			   long  pResSetStrtPt,
			   long  pNumRecReq,
			   char* pElemSetName,
			   char* pPrefRecSyn)
{ 
  DATA_DIR *lTop = NULL;
  DATA_DIR *lElm = NULL;
  DATA_DIR *lGen = NULL;
  boolean lError = false;
  
  /* Create a directory structure. */
  if ((lTop = dmake(tagPDUPrsntReq, ASN1_CONTEXT,20L)) == NULL)
    lError = true;

  /* Required reset set id, result set start point, and number of
     records requested. */
  if (lError ||
      (daddchar(lTop,tagGlobResSetId,ASN1_CONTEXT, 
		(CHAR*)pResSetId,safeStrlen(pResSetId)) == NULL) ||
      (daddnum(lTop,tagPrsntResSetStrtPt,ASN1_CONTEXT, 
	       (CHAR*)&pResSetStrtPt,sizeof(pResSetStrtPt)) == NULL) ||
      (daddnum(lTop,tagPrsntNumRecReq,ASN1_CONTEXT,
	       (CHAR*)&pNumRecReq,sizeof(pNumRecReq)) == NULL))
    lError = true;

  /* Optional element set names. */
  if (lError ||
      ((pElemSetName != NULL) &&
       (((lElm = daddtag(lTop,tagPrsntElmtSetNames,ASN1_CONTEXT)) == NULL) ||
	((lGen = daddtag(lElm,tagElmtSetNamesGen,ASN1_CONTEXT)) == NULL) ||
	((daddchar(lGen,tagElmtSetName,ASN1_CONTEXT,
		   (CHAR*)pElemSetName,safeStrlen(pElemSetName))) == NULL))))
    lError = true;

  /* Optional preferred record syntax. */
  if (lError ||
      ((pPrefRecSyn != NULL) &&
       (daddoid(lTop,tagSrchAndPrsntPrefRecSyn,
		ASN1_CONTEXT,pPrefRecSyn)) == NULL))
    lError = true;

#ifdef DEBUG
  if (!lError)
    BufDumpBer("Client PresentRequest",lTop);
#endif

  if (!lError)
    {
      /* Determine length, and build pBuf from directory. */
      rec_len(lTop);
      *pBufLen = asm_rec(lTop,(CHAR*)pBuf);
    }
  else *pBufLen = 0L;

  if (lTop != NULL) dfree(lTop);  
  return(!lError);
}
 


/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2PresentRespReadAPDU
 Purpose:    Read a present response PDU in BER from the buffer pBuf, and check
             for errors.  This function does not read the returned records or
             the diagnostics.
 Called by:  Z3950V2PresentRespReadVariantsAPDU
             Z3950V2PresentRespReadObjElemAPDU
 Parameters: pBuf          -- input;
             pNumRecRet    -- output;
             pNxtResSetPos -- output;
             pPrsntStat    -- output;
             pTop          -- output;
             pDiags        -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
Z3950V2PresentRespReadAPDU(char*      pBuf,
			   long*      pNumRecRet,
			   long*      pNxtResSetPos,
			   long*      pPrsntStat,
			   DATA_DIR** pTop,
			   list*      pDiags)
{
  DATA_DIR *lNxt = NULL;
  boolean lError = false;
 
  /* Check for errors in the present response PDU. */
  if (((*pTop = dalloc(20)) == NULL) ||
      (bld_dir((CHAR*)pBuf, *pTop) == false) ||
      ((lNxt = tagFindNext(*pTop,tagPDUPrsntResp)) == NULL))
    lError = true;

#ifdef DEBUG
  if (!lError)
    BufDumpBer("Client PresentResponse",*pTop);
#endif

  /* Begin decoding the present response sequence. */
  /* If any of the required arguments are NULL (unspecified)
     the pdu is non-conformant with the standard. */

  /* Read the number of records returned. */
  if (!lError &&
      ((lNxt = tagFindChild(lNxt,tagSrchAndPrsntNumRecRet)) != NULL))
    *pNumRecRet = dgetnum(lNxt);
  else lError = true;

  /* Read the next result set position. */
  if (!lError &&
      ((lNxt = tagFindNext(lNxt,tagSrchAndPrsntNxtResSetPos)) != NULL))
    *pNxtResSetPos = dgetnum(lNxt);
  else lError = true;

  /* Read the present status. */
  if (!lError &&
      ((lNxt = tagFindNext(lNxt,tagSrchAndPrsntPrsntStat)) != NULL))
    *pPrsntStat = dgetnum(lNxt);
  else lError = true;
  
  if (!lError)
   {
     /* Handle the error. */  
     /* Ensure present status is false, result count and number of records 
	returned is zero. */
     *pPrsntStat = tagPrsntStatFail;
     *pNxtResSetPos = *pNumRecRet = 0L;
   }

  return(!lError);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2PresentRespReadVariantsAPDU
 Purpose:    Read a present response PDU in BER from the buffer pBuff, and
             check for errors.  Reads the full PDU.  Expects to read records
             from the "VARIANT" element set.
 Called by:  Z3950V2PresentRespReadVariants
 Parameters: pBuf          -- input;
             pNumRecRet    -- output;
             pNxtResSetPos -- output;
             pPrsntStat    -- output;
             pRecs         -- output;
             pDiags        -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
Z3950V2PresentRespReadVariantsAPDU(char* pBuf, 
				   long* pNumRecRet, 
				   long* pNxtResSetPos, 
				   long* pPrsntStat,
				   list* pRecs,
				   list* pDiags)
{
  DATA_DIR *lTop = NULL;
  boolean lRetCode;

  lRetCode = 
    /* Read the buffer into lTop and decode. */
    (Z3950V2PresentRespReadAPDU(pBuf,pNumRecRet,pNxtResSetPos, 
				pPrsntStat,&lTop,pDiags) &&
     /* Decode the database or diagnostic records. */
     recPrsRecs(lTop,pDiags,VARIANT_ELEM_SET,pRecs));

  if (lTop != NULL) dfree(lTop);
  return(lRetCode);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   Z3950V2PresentRespReadObjElemAPDU
 Purpose:    Read a present response PDU in BER from the buffer pBuff, and
             check for errors.  Reads the full PDU.  Expects to read the full
             object element from the element set; this may be in the form of
             a full text, image, or multi-media object.
 Called by:  Z3950V2PresentRespReadObjElem
 Parameters: pBuf          -- input;
             pNumRecRet    -- output;
             pNxtResSetPos -- output;
             pPrsntStat    -- output;
             pElem         -- output;
             pObjElem      -- output;
             pObjElemLen   -- output;
             pTargetPtr    -- output;
             pDiags        -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
Z3950V2PresentRespReadObjElemAPDU(char*  pBuf, 
				  long*  pNumRecRet, 
				  long*  pNxtResSetPos, 
				  long*  pPrsntStat,
				  char** pElem,
				  char** pObjElem, 
				  long*  pObjElemLen,
				  char** pTargetPtr,
				  list*  pDiags)
{ 
  DATA_DIR *lTop = NULL;
  boolean lRetCode;

  /* Read the APDU. */
  lRetCode = 
    /* Read the buffer into lTop and decode. */
    (Z3950V2PresentRespReadAPDU(pBuf,pNumRecRet,pNxtResSetPos, 
				pPrsntStat,&lTop,pDiags) &&
     /* Decode the database or diagnostic records. */
     recPrsRecs(lTop,pDiags,OBJ_ELEM_ELEM_SET,
		pElem,pObjElem,pObjElemLen,pTargetPtr));

  if (lTop != NULL) dfree(lTop);
  return(lRetCode);
}

