/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Module:   connV2Cli.c

 Purpose:  This module provides the functional interface between the client
           API (see appConnCli.c) and the Z39.50-V2 calls (see cli.c).
---------------------------------------------------------------------------*/

#include "cutil.h"
#include "list.h"
#include "appConnMgr.h"

#include "connV2Cli.h"
#include "berutil.h"
#include "attr.h"
#include "oid.h"
#include "tag.h"
#include "cli.h"



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2CliPduInfo
 Purpose:    Reads the data on the connection and determines the pdu type
             and the length of the packet.  The data read is stored in the
             connection's buffer.
             Note that this function may return a status of 
             appConnStatusNotFinished, which indicates that not all of the
             data has yet been received and parsed.  The client should 
             continue to call the function until it returns either an error
             or non-error status.  This is designed to allow the client
             application to process other events and user input while
             waiting for the response to arrive.
 Called by:  appConnPduInfo
 Parameters: conn         -- input;
             pduType      -- output;
             packetLength -- output;
             diags        -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus 
z3950V2CliPduInfo(appConn* conn, 
		  long*    pduType,
		  long*    packetLength,
		  list*    diags)
{
  appConnV2 *zConn = (appConnV2*)conn;
  appConnStatus as;
  int pduTag = 0L;

  as = z3950V2PduInfo(conn,packetLength,diags);

  if (as == appConnStatusOK)
   {
     /* Grab the Pdu tag */
     get_tag(&pduTag,(CHAR*)conn->buf,zConn->totalLength);

     /* Check for a retrieval request inside a search APDU. */
     switch (pduTag)
      { 
      case tagPDUInitResp: 
	*pduType = appConnInitResp; break;

      case tagPDUSrchResp:
	/* For backward compatibility with 88, retrieval can be performed
	   with a search.  That means we have to remember what we were
	   doing so that we can know what to expect
	   */
       { 
	 if (zConn->doingRetrieval)
	   *pduType = appConnRetrieveResp;
	 else *pduType = appConnSearchResp;
	 break;
       }

      case tagPDUPrsntResp:
       {
	 if (zConn->doingRetrieval)
	   *pduType = appConnRetrieveResp;
	 else *pduType = appConnNextHeadlinesResp;
	 break;
       }

      default: *pduType = appConnUnknown;
      }
   }

  return(as);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2InitReqWrite
 Purpose:    Write an initialization request message to the connection, conn.
             For this release, authentication (user, password, clientInfo) and 
             userInfo are ignored.
 Called by:  appConnInitReqWrite
 Parameters: conn       -- input;
             user       -- input;
             password   -- input;
             clientInfo -- input;
             userInfo   -- ignored;
             diags      -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2InitReqWrite(appConn* conn,
		    char*    user,    
		    char*    password,
		    char*    clientInfo,
		    any*     userInfo,
		    list*    diags)
{
  long size = conn->realBufSz;
  char *message;

  /* Create an init message in conn->buf of length = size. */
  if (Z3950V2InitReqWriteAPDU(conn->buf,&size,size,size,NULL,
			      Z3950V2_IMP_ID,Z3950V2_IMP_NAME,Z3950V2_IMP_VER))
   {
     /* Send buf to remote system and create a diagnostic if a network
	problem was encountered. */
     z3950V2WriteMessage(conn->net,conn->buf,size,diags);
   }

  else
   {
     /* Unable to create init request; create & return diagnostic. */
     appConnDiag *diag = appConnDiagMake(appConnDiagClientError,
					 "unable to create init request");
     *diags = collecting(*diags,diag);
   }

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2InitRespRead
 Purpose:    Reads an initialization response packet from the connection.
             If the connection is refused, a list of explanatory diagnostics
             is returned in diags.
 Called by:  appConnInitRespRead
 Parameters: conn     -- input;
             userInfo -- output;
             diags    -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2InitRespRead(appConn* conn,  
		    any**    userInfo,
		    list*    diags)
{ 
  boolean result = false;
  long prefMsgSiz, maxRecSiz; 
  char *impId = NULL;
  char *impName = NULL;
  char *impVer = NULL; 
  appConnDiag *diag = NULL;

  /* Read and decode initialization response message. */
  if (Z3950V2InitRespReadAPDU(conn->buf,&prefMsgSiz,&maxRecSiz,&result,
			      &impId,&impName,&impVer,diags))

   {
     if (result)
      {
	/* Set the negotiated buffer size. */
	conn->negotiatedBufSz = prefMsgSiz;

	/* Save server information in variable userInfo. */
	if (userInfo != NULL)
	 { 
	   char *msg;
	   if ((msg = (char*)safeMalloc(sizeof(char)*
					(33L+safeStrlen(impId)
					+safeStrlen(impName)
					+safeStrlen(impVer))))
	       != NULL)
	    { 
	      sprintf(msg,"Id:%s Name:%s Ver:%s",impId,impName,impVer);
	      *userInfo = stringToAny(msg);
	      safeFree(msg);
	    }
	 }
      }
     else
      {
	/* No negotiation; set negotiated buffer size to zero. */
	conn->negotiatedBufSz = 0L;

	/* Initialization failed. Create a diagnostic. */
	diag = appConnDiagMake(appConnDiagAccessDenied,"");
	*diags = collecting(*diags,diag);
			    
      }
   }
  else
   { 
     /* Unable to parse init response; create & return diagnostic. */
     diag = appConnDiagMake(appConnDiagBadPacket,
			    "unable to parse init response");
     *diags = collecting(*diags,diag);
   }

  /* Free up memory. */
  safeFree(impId);
  safeFree(impName);
  safeFree(impVer);
  
  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2SearchReqWrite
 Purpose:    Writes a search request packet to the connection.
             In this implementation, maxHits will be used as the large set 
             upper bound will most likely be used by the server to limit the
             search to a result set of maxHits.
             Small set lower bound: "If the number of hits is less than or
             equal to the small set upper bound, all records are to be
             returned in the search Response, composed as specified by the
             small set element set names."
             Large set upper bound: "If the number of hits is equal to or
             greater than the large set upper bound, no records are to be
             returned."
             Medium set present number: "Medium set present number is the
             maximum number of records to be returned in the search response
             if the number of hits is between large set lower bound and small
             set upper bound, composed as specified by medium set record
             element set names."
 Called by:  appConnSearchReqWrite
 Parameters: conn          -- input;
             maxHits       -- input;
             hitsRequested -- input;
             rankingScheme -- ignored;
             dbName        -- input;
             qry           -- input;
             userInfo      -- ignored;
             diags         -- output;
 Returns:    appConnStatus  
---------------------------------------------------------------------------*/
appConnStatus
z3950V2SearchReqWrite(appConn*    conn,  
		      long        maxHits,
		      long        hitsRequested, 
		      long        rankingScheme, 
		      char*       dbName, 
		      appConnQry* qry,
		      any*        userInfo,
		      list*       diags)
{ 
  appConnV2 *zConn = (appConnV2*)conn;
  long size;

  if (maxHits < 0) maxHits = MAXLONG;
  if (hitsRequested < 0) hitsRequested = MAXLONG;

  if (Z3950V2SearchReqWriteAPDU(conn->buf,&size,0L,maxHits,hitsRequested,
				true,zConn->defResSetName,dbName,
				zConn->defElmtSetName,zConn->defElmtSetName,
				zConn->defPrefRecSyn,qry))
    { 
      /* Send it to the remote system. */
      z3950V2WriteMessage(conn->net,conn->buf,size,diags);

    }
  else
   {
     /* Unable to create search request; create and return diagnostic. */
     appConnDiag *diag = appConnDiagMake(appConnDiagClientError,
					 "unable to create search request");
     *diags = collecting(*diags,diag);
   }

  /* Remember we are not doing a retrieval operation. */
  zConn->doingRetrieval = false;

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2SearchRespRead
 Purpose:    Reads a search response packet from the connection.  This 
             function reads data from the connection and parses it into
             hits.  The total number of hits in the result set is returned
             in the parameter totalHits; this value may be greater than or
             equal to the actual number of hits returned (i.e. the length
             of the hits list).
 Called by:  appConnSearchRespRead
 Parameters: conn         -- input;
             totalHits    -- output;
             qryTermsUsed -- output;
             hits         -- output;
             userInfo     -- ignored;
             diags        -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2SearchRespRead(appConn* conn,
		      long*    totalHits, 
		      list*    qryTermsUsed,
		      list*    hits,
		      any**    userInfo,
		      list*    diags)
{ 
  long numHitsRet, nxtResSetPos;
  boolean srchStat;
  long resSetStat, prsntStat;
  list hitsOrDiags;

  if (!Z3950V2SearchRespReadVariantsAPDU(conn->buf,totalHits,
					 &numHitsRet,&nxtResSetPos, 
					 &srchStat,&resSetStat,&prsntStat,
					 hits,diags))
   {
     /* Unable to parse search response; create & return diagnostic. */
     appConnDiag *diag = 
       appConnDiagMake(appConnDiagBadPacket,
		       "unable to parse search response");
     *diags = collecting(*diags,diag);
   }

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2NextHeadlinesReqWrite
 Purpose:    Writes a next headlines request packet to the connection.  This
             function requests more hits from the server.  Hits beginning at
             (startingFromHit) to (startingFromHit + hitsRequested - 1), 
             inclusive, are requested.
 Called by:  appConnNextHeadlinesReqWrite
 Parameters: conn            -- input;
             startingFromHit -- input;
             userInfo        -- ignored;
             diags           -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2NextHeadlinesReqWrite(appConn* conn,
			     long     startingFromHit, 
			     long     hitsRequested,
			     any*     userInfo,
			     list*    diags)
{ 
  appConnV2 *zConn = (appConnV2*)conn;
  long size;

  if (Z3950V2PresentReqWriteAPDU(conn->buf,&size,zConn->defResSetName,
				 startingFromHit,hitsRequested,
				 zConn->defElmtSetName,zConn->defPrefRecSyn))
   {
     /* Send buf to remote system and create a diagnostic if a network
	problem was encountered. */
     z3950V2WriteMessage(conn->net,conn->buf,size,diags);
   }

  else
   {
     /* Unable to create present request; create & return diagnostic. */
     appConnDiag *diag = 
       appConnDiagMake(appConnDiagClientError,
		       "unable to create present request");
     *diags = collecting(*diags,diag);
   }

  zConn->doingRetrieval = false;

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2NextHeadlinesRespRead
 Purpose:    Reads a next headlines response packet from the connection.
 Called by:  appConnNextHeadlinesRespRead
 Parameters: conn     -- input;
             hits     -- output;
             userInfo -- ignored;
             diags    -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/

appConnStatus z3950V2NextHeadlinesRespRead(appConn* conn,
					   list*    hits,
					   any**    userInfo,
					   list*    diags)
{ 
  long numHitsRet, nxtResSetPos, prsntStat;

  if (!Z3950V2PresentRespReadVariantsAPDU(conn->buf,&numHitsRet,&nxtResSetPos, 
					  &prsntStat,hits,diags))
   { 
     /* Unable to read the present response, create a diagnostic. */
     appConnDiag *diag = 
       appConnDiagMake(appConnDiagBadPacket,
		       "unable to parse present response");
     *diags = collecting(*diags,diag);
   }

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2RetrieveReqWrite
 Purpose:    Frees up old retrieval information saved in the zConn.
 Called by:  z3950V2RetrieveReqWrite
 Parameters: zConn -- input; a handle to a Z39.50 connection.
 Returns:    void
---------------------------------------------------------------------------*/
static void
z3950V2FreeOldRetrievalRequest(appConnV2* zConn)
{
  freeAny(zConn->recIdReq); zConn->recIdReq = NULL;
  safeFree(zConn->docIdReq); zConn->docIdReq = NULL;
  safeFree(zConn->elemIdReq); zConn->elemIdReq = NULL;
  safeFree(zConn->varIdReq); zConn->varIdReq = NULL;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2RetrieveReqWrite
 Purpose:    Writes a retrieval request packet to the connection.  Retrieval
             may be performed using one of three methods.  The first method
             of retrieval is to select a hit from a previous search; this
             retrieval is specified by the hit's position in the result list.
             The second method is to specify the document identifier of the
             requested document, and the third is to specify the ULx (Uniform
             Resource Locator or Name) of the requested document.  In all 
             cases, if the document is composed of multiple components,
             or elements, (such as an citation, an abstract, a table of 
             contents, chapters, sections, etc.), an element specifier may
             be specified.  If an element is available in multiple variants
             (such as different languages, formats, etc.), a variant 
             identifier may also be specified.  If no element identifier or
             variant identifier is specified, a default element and variant
             is selected by the server.
             startByte and lengthInBytes are ignored. 
             The server returns as much as it wants.
 Called by:  appConnRetrieveReqWrite
 Parameters: conn          -- input;
             retrievalType -- input;
             resSetPos     -- input;
             recId         -- input;
             docId         -- input;
             elemId        -- input;
             varId         -- input;
             dbName        -- input;
             startByte     -- ignored;
             lengthInBytes -- ignored;
             userInfo      -- ignored;
             diags         -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2RetrieveReqWrite(appConn* conn,
			long     retrievalType,
			long     resSetPos,
			any*     recId,
			char*    docId,
			char*    elemId,
			char*    varId,
			char*    dbName,
			long     startByte,
			long     lengthInBytes,
			any*     userInfo,
			list*    diags)
{ 
  appConnV2 *zConn = (appConnV2*)conn;
  long size;
  char *elemSetName = NULL;
  char *diagStr = NULL;

  /* Check if this record had just been requested. Compare retrieval type
     with the previous retrieval type, with the same resSetPos, recId,
     docId, elemId, and varId, with the previous request.  If they are
     the same, assume client is asking for next piece. */

  /* Ask for a single record. */
  switch (retrievalType) 
   { 
   case appConnRetrieveByURx:
    { 
      appConnQry lQry;
      lQry.type = qryLeafDocId;
      lQry.term = stringToAny(docId);

      /* Retrieval by Universal Resource Identifier. */
      if ((zConn->retrievalTypeReq == retrievalType) &&
	  !strcmp(zConn->docIdReq,docId) &&
	  !strcmp(zConn->elemIdReq,elemId) &&
	  !strcmp(zConn->varIdReq,varId) &&
	  (zConn->targetPtr != NULL))

	/* Ask for the next piece. */
	elemSetName = zConn->targetPtr;

      else 
       { 

	 /* Asking for retrieval of a new variant. */
	 z3950V2FreeOldRetrievalRequest(zConn);

	 /* Remember what we are requesting. */
	 zConn->retrievalTypeReq = retrievalType;
	 zConn->docIdReq = safeStrdup(docId);
	 zConn->elemIdReq = safeStrdup(elemId);
	 zConn->varIdReq = safeStrdup(varId);

	 /* Ask for the first piece by using an element set name that is the 
	    same as the variant idenfifier. */
	 elemSetName = varId;
       }
      
      if (!Z3950V2SearchReqWriteAPDU(conn->buf,&size,0L,1L,1L,
				     true,zConn->defResSetName,dbName,
				     elemSetName,elemSetName,
				     zConn->defPrefRecSyn,&lQry))
	/* If not successful, create a diagnostic string. */
	diagStr = "unable to create URL search request";
	
      freeAny(lQry.term);
      break;
    }

   case appConnRetrieveByRecId:
    { 
      appConnQry lQry;
      lQry.type = qryLeafRecId;
      lQry.term = recId;
 
      /* Retrieval by Local Record Identifier. */
      if ((zConn->retrievalTypeReq == retrievalType) &&
	  anycmp(zConn->recIdReq,recId) &&
	  !strcmp(zConn->elemIdReq,elemId) &&
	  !strcmp(zConn->varIdReq,varId) &&
	  (zConn->targetPtr != NULL))

	/* Ask for the next piece. */
	elemSetName = zConn->targetPtr;

      else 
       {
	 /* Asking for retrieval of a new variant. */
	 z3950V2FreeOldRetrievalRequest(zConn);

	 /* Remember what we are requesting. */
	 zConn->retrievalTypeReq = retrievalType;
	 zConn->recIdReq = duplicateAny(recId);
	 zConn->elemIdReq = safeStrdup(elemId);
	 zConn->varIdReq = safeStrdup(varId);

	 /* Ask for the first piece by using an element set name that is the 
	    same as the variant idenfifier. */
	 elemSetName = varId;
       }

      if (!Z3950V2SearchReqWriteAPDU(conn->buf,&size,0L,1L,1L,
				     true,zConn->defResSetName,dbName,
				     elemSetName,elemSetName,
				     zConn->defPrefRecSyn,&lQry))
	/* If not successful, create a diagnostic string. */
	diagStr = "unable to create RecId search request";

      break;
    }

   case appConnRetrieveByPosition:
    { 
      /* Retrieval by Local Record Identifier. */
      if ((zConn->retrievalTypeReq == retrievalType) &&
	  (zConn->resSetPosReq == resSetPos) &&
	  !strcmp(zConn->elemIdReq,elemId) &&
	  !strcmp(zConn->varIdReq,varId) &&
	  (zConn->targetPtr != NULL))

	/* Ask for the next piece. */
        elemSetName = zConn->targetPtr;

      else 
       {
	 /* Asking for retrieval of a new variant. */
	 z3950V2FreeOldRetrievalRequest(zConn);

	 /* Then, remember what we are requesting. */
	 zConn->retrievalTypeReq = retrievalType;
	 zConn->resSetPosReq = resSetPos;
	 zConn->elemIdReq = safeStrdup(elemId);
	 zConn->varIdReq = safeStrdup(varId);

	 /* Ask for the first piece by using an element set name that is the 
	    same as the variant idenfifier. */
	 elemSetName = varId;
       }

      /* Retrieval by position from a result set. */
      if (!Z3950V2PresentReqWriteAPDU(conn->buf,&size,
				      zConn->defResSetName,resSetPos,1L,
				      elemSetName,zConn->defPrefRecSyn))
	/* If not successful, create a diagnostic string. */
	diagStr = "unable to create present request";
      
      break;
    }
   }

  /* Record that we are doing a retrieval, and not a search. */
  zConn->doingRetrieval = true;

  /* Free up the old target pointer, if there is one. */
  safeFree(zConn->targetPtr);
  zConn->targetPtr = NULL;

  /* Send buf to remote system and create a diagnostic if a network
     problem was encountered. */
  if (diagStr == NULL)
    z3950V2WriteMessage(conn->net,conn->buf,size,diags);
  /* Create a diagnostic. */
  else 
    *diags = collecting(*diags,appConnDiagMake(appConnDiagBadPacket,diagStr));

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2RetrieveRespRead
 Purpose:    Reads a retrieval response packet from the connection.  A 
             diagnostic error code of appConnDiagRequestOutOfRange is 
             returned in the list of diagnostics if the data returned
             contains the end of the document.
 Called by:  appConnRetrieveRespRead
 Parameters: conn       -- input;
             data       -- output;
             dataLength -- output;
             userInfo   -- ignored;
             diags      -- output;
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus 
z3950V2RetrieveRespRead(appConn* conn,
			char**   data,
			long*    dataLength,
			any**    userInfo,
			list*    diags)
{
  appConnV2 *zConn = (appConnV2*)conn;
  long totalHits, numHitsRet, nxtResSetPos;
  boolean srchStat;
  long resSetStat, prsntStat;
  char *elem;
  char *diagStr = NULL;

  switch (zConn->retrievalTypeReq)
   { 
   case appConnRetrieveByURx:
   case appConnRetrieveByRecId:
    { 
      if (!Z3950V2SearchRespReadObjElemAPDU(conn->buf,&totalHits,
					    &numHitsRet,&nxtResSetPos, 
					    &srchStat,&resSetStat,&prsntStat,
					    &elem,data,dataLength,
					    &zConn->targetPtr,diags))
	/* If not successful, create a diagnostic string. */
	diagStr = "unable to parse search request";

      break;
    }

   case appConnRetrieveByPosition:
    { 
      if (!Z3950V2PresentRespReadObjElemAPDU(conn->buf,&numHitsRet,
					     &nxtResSetPos,&prsntStat,
					     &elem,data,dataLength,
					     &zConn->targetPtr,diags))
	/* If not successful, create a diagnostic string. */
	diagStr = "unable to parse present request";

      break;
    }
   }

  /* Free up the element set name. */
  safeFree(elem);

  /* Check for a diagnostic, and create one if necessary. */
  if (diagStr != NULL)
    *diags = collecting(*diags,appConnDiagMake(appConnDiagBadPacket,diagStr));

  /* Check to see if targetPtr was NULL.  If so, return a diagnostic
     to say that we've finished retrieving the whole variant. */
  else if (zConn->targetPtr == NULL)
    *diags = collecting(*diags,
			appConnDiagMake(appConnDiagRequestOutOfRange,
					"end of variant"));

  return(appConnStatusOK);
}

