/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Module:   connV2Mgr.c

 Purpose:  This module provides a set of shared functions used by both the
           client (connV2Cli.c and server (connV2Srv.c) modules.
---------------------------------------------------------------------------*/

#include "cutil.h"
#include "appConnMgr.h"

#include "oid.h"
#include "berutil.h"
#include "connV2Mgr.h"


/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2WriteMessage
 Purpose:    Writes buff to the network connection and checks for write errors.
 Called by:  z3950V2InitReqWrite, z3950V2SearchReqWrite,
             z3950V2NextHeadlinesReqWrite, z3950V2RetrieveReqWrite
 Parameters: net   -- input; a handle to the nework connection, nConn.
             buf   -- input; a handle to the data to be written.
             size  -- input; a long integer whose value is the number of
                      bytes that need to be written.
             diags -- output; a handle to a list of diagnostics,
                      where each item in the list is of type appConnDiag
                      (see ../appConnMgr.c).  If the operation was 
                      unsuccessful, a list of diags are returned in
                      this parameter.
 Returns:    nConnStatus
---------------------------------------------------------------------------*/
nConnStatus
z3950V2WriteMessage(nConn* net,
		    char*  buf,
		    long   size,
		    list*  diags)
{
  long i;
  nConnStatus res;

  /* send it to the remote server, and wait for a response */
  for (i = 0; i < size; i++)
   { res = nConnWrite(net,buf[i]);
     if (res != nConnStatusOK)
       return(res);
   }

  res = nConnFlush(net);		

  if (res != nConnStatusOK)
    *diags
      = collecting(*diags,
		   appConnDiagMake(appConnDiagConnErrWrite,
				   "error writing %ld bytes to connection",
				   size));
  return(res);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2Open
 Purpose:    Open a Z39.50-V2 connection.
 Called by:  appConnOpen
 Parameters: conn        -- output; a pointer to the handle of the
                            appConn. It holds information about the state
                            of the connection; created by this function.
             name        -- input; "Z3950V2".
             net         -- input; a handle to the network connection nconn.
                            The network connection is created by the 
                            nConnOpen() function (see nConnMgr.h).
             buffer      -- input; a character buffer used to hold the 
                            incoming and outgoing data to be transmitted.
                            It is reused over the lifetime of the connection.
                            Must be created by caller.
             bufSz       -- input; a long integer whose value is the size of
                            the buffer.
             timeOutSecs -- input; a long integer whose value is the timeout
                            time for a transaction.
             numArgs     -- ignored.
 Returns:    nConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2Open(appConn** conn,
	    char*     connType,
	    nConn*    net,
	    char*     buffer,
	    long      bufSz,
	    long      timeOutSecs,
	    long      numArgs,
	    ...)
{
  appConn *c;
  appConnV2 *zConn;

  *conn = c = (appConn*)safeMalloc(sizeof(appConnV2));

  c->name = safeStrdup(connType);
  c->net = net;
  c->buf = buffer;
  c->realBufSz = bufSz;
  c->negotiatedBufSz = bufSz;
  c->timeOutSecs = timeOutSecs;

  /* Set up the connection's default result set id, element set name, 
     preferred record syntax. */
  zConn = (appConnV2*)(*conn);
  zConn->defResSetName = safeStrdup(DEF_RES_SET_NAME);
  zConn->defElmtSetName = safeStrdup(VARIANT_ELEM_SET_NAME);
  zConn->defPrefRecSyn = safeStrdup(OID_RECSYN_GENERIC);

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2Free
 Purpose:    Frees the z3950V2 appConn, and any memory related to the 
             implementation-specific pieces of the protocol.  
             Note that this does not free the network connection nConn.
 Called by:  appConnFree
 Parameters: conn -- input; a handle to an appConn. It holds information
                     about the state of the connection; it is freed by this
                     function.
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2Free(appConn* conn)
{
  appConnV2 *zConn = (appConnV2*)conn;

  safeFree(zConn->defResSetName);
  safeFree(zConn->defElmtSetName);
  safeFree(zConn->defPrefRecSyn);
  freeAny(zConn->recIdReq);
  safeFree(zConn->elemIdReq);
  safeFree(zConn->varIdReq);
  safeFree(zConn->targetPtr);
  safeFree(zConn->refId);
  safeFree(zConn->dbName);

  safeFree(zConn);

  return(appConnStatusOK);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   z3950V2PduInfo
 Purpose:    Reads data from the connection, and returns the length of the
             packet.  The data read is stored in the connection's buffer.
             Note that this function may return a status of 
             appConnStatusNotFinished, which indicates that not all of the
             data has yet been received and parsed.
 Called by:  z3950V2CliPduInfo, z3950V2SrvPduInfo
 Parameters: conn         -- input; a handle to an appConn. It holds
                             information about the state of the connection;
                             may be modified by this function.
             packetLength -- output; a handle to a long integer whose value
                             is the length of the received packet, or 
                             appConnUnknownPacketLength (-1) if the length is
                             unknown.
             diags        -- output; a handle to a list of diagnostics,
                             where each item in the list is of type 
                             appConnDiag (see ../appConnMgr.c).  If the
                             operation was unsuccessful, a list of diags
                             are returned in this parameter.
 Returns:    appConnStatus
---------------------------------------------------------------------------*/
appConnStatus
z3950V2PduInfo(appConn* conn, 
	       long*    packetLength,
	       list*    diags)
{
  appConnV2 *zConn = (appConnV2*)conn;
  time_t startTime = time(NULL);
  long i, remainder = 0;
  
  if (packetLength != NULL)
    *packetLength = appConnUnknownPacketLength; /* the default */ 

  if (zConn->readSoFar == zConn->totalLength)
   { /* we finished reading the previous batch, and are being asked to 
	read a new batch, so we need to initialize things
      */
     zConn->totalLength = 0;
     zConn->readSoFar = 0;
     /* Clear out the first bytes of the buffer, so there's no ambiguity
        about the length of the pdu. Is CompleteBER may be broken because
	is may return a bogus remainder if not all the length bits are
        present. Don't know if it's broken for indefinite lengths. */
     memset(conn->buf,0L,10L);
   }

  /* read the rest of the message if necessary */
  while (!IsCompleteBER((CHAR*)conn->buf, zConn->readSoFar, &remainder)) 
   { if (packetLength != NULL) *packetLength = remainder;
     for (i = 0; i < (remainder?remainder:1); i++)
      { unsigned char nextChar;
	nConnStatus err;
	err = nConnRead(conn->net,&nextChar);
	if (err == nConnStatusOK)
	 { conn->buf[zConn->readSoFar] = nextChar;
	   zConn->readSoFar++;
	 }
	else if (err == nConnStatusReadFailed)
	 { *diags = collecting(*diags,
			       appConnDiagMake(appConnDiagConnErrRead,""));
	   return(appConnStatusOK);
	 }
	else if (err == nConnStatusReadEOF)
	 { *diags = collecting(*diags,
			       appConnDiagMake(appConnDiagConnClosed,""));
	   return(appConnStatusOK);
	 }
	else if (err == nConnStatusReadNoData)
	  ; /* no data yet, try again later */
	if (timeOut(startTime,conn->timeOutSecs))
	  return(appConnStatusNotFinished);
      }
   }

  /* Set the actual length of the packet. */
  zConn->totalLength = remainder;

  /* Return status */
  return(appConnStatusOK);
}

