/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Module:   recPrs.c

 Purpose:  This module provides functions for parsing the records field of a
           Z39.50 search or present response.  More specifically, it parses
           Generic Record Syntax, and any diagnostics or surrogate diagnostics
           that may also appear.  The parser could easily be extended to parse
           MARC and SUTRES records.  The top-level function is recPrsRecs().
---------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include "cutil.h"
#include "list.h"
#include "appConnMgr.h"

#include "connV2Mgr.h"
#include "oid.h"
#include "berutil.h"
#include "tag.h"
#include "diag.h"
#include "recPrs.h"



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsVarTriple
 Purpose:    Parses a variant (triple) in GRS.
 Called by:  recPrsGrsMetaTargetPtr, recPrsGrsMetaDataElem
 Parameters: pNode  -- input;
             pClass -- output;
             pType  -- output;
             pValue -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsVarTriple(DATA_DIR* pNode, 
		   long*     pClass, 
		   long*     pType, 
		   void**    pValue)
{ 
  DATA_DIR *lNode = NULL;
  DATA_DIR *lTmp = NULL;

  /* Sequence */
  if ((lNode = pNode->ptr.child) == NULL)
    return(false); 

  /* Class */
  if ((lTmp = tagFindNext(lNode,tagGRSVarTriplesClass)) != NULL)
    *pClass = dgetnum(lTmp);
  else return(false);
     
  /* Type */
  if ((lTmp = tagFindNext(lNode,tagGRSVarTriplesType)) != NULL)
    *pType = dgetnum(lTmp);
  else return(false);
	      
  /* Value is a choice. */
  if (((lTmp = tagFindNext(lNode,tagGRSVarTriplesValue)) != NULL) &&
      ((lTmp = lTmp->ptr.child) != NULL))
      switch (lTmp->fldid)
       { case tagUnivInt:
	 case tagUnivBool:
	  { *pValue = (void*)dgetnum(lTmp); 
	    break;
	  }
	 case tagUnivOctStr:
	 case tagUnivObjId:
	 case tagUnivGenTime:
	 case tagUnivVisStr:
	 case tagUnivGenStr:
	  { *pValue = safeStrndup((char*)lTmp->ptr.data,1L+lTmp->count);
	    break;
	  }
	 case tagUnivExt:
	  break;
	 case tagUnivNull:
	  { *pValue = NULL;
	    break;
	  }
	 default: 
	  { return(false); 
	    break;
	  }
	 }
  else return(false);
  
  return(true);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsTagType
 Purpose:    Parses the tag type in GRS.
 Called by:  recPrsGrsObjElem, recPrsGrsVariant
 Parameters: pDir     -- input;
             pTagType -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsTagType(DATA_DIR* pDir,
		 long*     pTagType)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp = NULL;

  /* Find the tag type in a tagged element of a GRS record. 
     Tag Type is always an integer.  */
  if ((lTmp = tagFindNext(pDir,tagGRSTagType)) != NULL)
    *pTagType = dgetnum(lTmp);

  else lRetCode = false;

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsTagVal
 Purpose:    Parses the tag value in GRS.
 Called by:  recPrsGrsObjElem, recPrsGrsVariant
 Parameters: pDir        -- input;
             pTagValType -- output;
             pTagValNum  -- output;
             pTagValStr  -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsTagVal(DATA_DIR* pDir,
		long*     pTagValType,
		long*     pTagValNum,
		char**    pTagValStr)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp = NULL;
  DATA_DIR *lTagValDir = NULL;

  /* Find the tag value, an integer or a string. */
  if ((lTmp = tagFindNext(pDir,tagGRSTagVal)) != NULL)

   { 
     /* Check for an integer tag. */
     if ((lTagValDir = tagFindChild(lTmp,tagGRSTagValNum)) != NULL)
      { 
	/* Copy the integer. */
	*pTagValNum = dgetnum(lTagValDir); 

	/* The type of the value is numeric. */
	*pTagValType = tagGRSTagValNum;
      }

     /* Check for a string tag. */
     else if ((lTagValDir = tagFindChild(lTmp,tagGRSTagValStr)) != NULL)

      { 
	/* Copy the string. */
	*pTagValStr = safeStrndup((char*)lTagValDir->ptr.data,
				  1L+lTagValDir->count);
	
	/* The type of the value is a string. */
	*pTagValType = tagGRSTagValStr;
      }

     else lRetCode = false;
   }

  else lRetCode = false;

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsCntsInt
 Purpose:    Parses an integer in GRS.  Can come in many different datatypes, 
             but only one is valid (choice).  Limitation of the BER routines
             is that there is no way to distinguish the class without looking
             at the underlying data structure.
 Called by:  recPrsGrsVariant
 Parameters: pDir -- input;
             pVal -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsCntsInt(DATA_DIR* pDir,
		 long*     pVal)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp;

  if (((lTmp = tagFindNext(pDir,tagGRSContent)) != NULL) &&
      ((lTmp = lTmp->ptr.child) != NULL) &&
      (lTmp->Class == ASN1_UNIVERSAL) &&
      (lTmp->fldid == tagUnivInt))

    *pVal = dgetnum(lTmp);

  else lRetCode = false;

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsCntsStr
 Purpose:    Parses a string in GRS.
 Called by:  recPrsGrsVariant
 Parameters: pDir -- input;
             pVal -- output;
             pLen -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsCntsStr(DATA_DIR* pDir,
		 char**    pVal,
		 long*     pLen)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp;

  if (((lTmp = tagFindNext(pDir,tagGRSContent)) != NULL) &&
      ((lTmp = lTmp->ptr.child) != NULL) &&
      (lTmp->Class == ASN1_UNIVERSAL) &&
      (((lTmp->fldid == tagUnivVisStr)) ||
       ((lTmp->fldid == tagUnivGenStr)) ||
       ((lTmp->fldid == tagUnivOctStr))))

   { 
     *pLen = lTmp->count;
     *pVal = safeStrndup((char*)lTmp->ptr.data,1L+*pLen);
   }

  else lRetCode = false;

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsCntsDate
 Purpose:    Parses a date in GRS.
 Called by:  recPrsGrsVariant
 Parameters: pDir -- input;
             pVal -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsCntsDate(DATA_DIR* pDir,
		  date**    pVal)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp;

  if (((lTmp = tagFindNext(pDir,tagGRSContent)) != NULL) &&
      ((lTmp = lTmp->ptr.child) != NULL) &&
      (lTmp->Class == ASN1_UNIVERSAL) &&
      ((lTmp->fldid == tagUnivGenTime)))

    *pVal = YYYYMMDDHHMMSSToDate((char*)lTmp->ptr.data);

  else lRetCode = false;

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsMetaDataElem
 Purpose:    Parses the meta data in GRS.  Note that this function does not
             yet handle Series Order, Usage Right, hits, displayName, message.
 Called by:  recPrsGrsVariant
 Parameters: pDir  -- input;
             pTag  -- input;
             pElem -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsMetaDataElem(DATA_DIR*        pDir,
		      char*            pTag,
		      appConnElement** pElem)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp;
  
  if (((lTmp = tagFindNext(pDir,tagGRSMetaData)) != NULL) &&
      ((lTmp = tagFindChild(lTmp,tagGRSElmtMetaDataSupVariants)) !=NULL) &&
      ((lTmp = tagFindChild(lTmp,tagUnivSeq)) != NULL))

    /* Determine all the variants. */
   { list lVariants = NULL;

     while (lTmp != NULL)

      { 
	appConnVariant *lVariant = NULL;
	DATA_DIR *lNodeVar = NULL; /* Assume Variant-1 is being used. */
	char *lName        = NULL; /* Name of the Variant. */
	char *lComposition = NULL; /* Media Type */
	long lSize         = 0L;   /* Size of variant in bytes */

	if (((lNodeVar = tagFindChild(lTmp,tagGRSVarTriples)) != NULL) &&
	    ((lNodeVar = tagFindChild(lNodeVar,tagUnivSeq)) != NULL))

	  /* Determine the set of triples for each variant. */
	 { while (lNodeVar != NULL)
	    { long  lClass = 0L;
	      long  lType = 0L;
	      void *lValue = NULL;
	      
	      recPrsGrsVarTriple(lNodeVar,&lClass,&lType,&lValue);

	      switch (lClass)
	       { case tagVarClassVarId:
		   lName = lValue; break;
		 case tagVarClassBodyPartType:
		   lComposition = lValue; break;
		 case tagVarClassMedaDataRet:
		  { switch (lType)
		     { case tagVarClassMedaDataRetSize:
			 lSize = (long)lValue; break;
		       default:
			 break;
		     }
		  }
		 default:
		   break;
	       }

	      lNodeVar = lNodeVar->next;
	    }
	 }

	lVariant = appConnMakeVariant(lName,lComposition,NULL,NULL,
				      lSize,0L,0L);
	lVariants = collecting(lVariants, lVariant);
	
	safeFree(lName);
	safeFree(lComposition);

	lTmp = lTmp->next;
      }

     *pElem = appConnMakeElement(pTag,NULL,lVariants,NULL);
   }

  else lRetCode = false;
  
  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsMetaTargetPtr
 Purpose:    Parses the target pointer in GRS.
 Called by:  recPrsGrsObjElem
 Parameters: pDir       -- input;
             pTargetPtr -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsMetaTargetPtr(DATA_DIR* pDir,
		       char**    pTargetPtr)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp = NULL;

  if (((lTmp = tagFindNext(pDir,tagGRSMetaData)) != NULL) &&
      ((lTmp = tagFindChild(lTmp,tagGRSElmtMetaDataSupVariants)) != NULL) &&
      ((lTmp = tagFindChild(lTmp,tagUnivSeq)) != NULL))

    /* Determine all the variants. */
   { 
     while (lTmp != NULL)
      { 
	DATA_DIR *lNodeVar = NULL; /* Assume Variant-1 is being used. */

	if (((lNodeVar = tagFindChild(lTmp,tagGRSVarTriples)) != NULL) &&
	    ((lNodeVar = tagFindChild(lNodeVar,tagUnivSeq)) != NULL))

	  /* Determine the set of triples for each variant. */
	 { 
	   while (lNodeVar != NULL)
	    { 
	      long  lClass = 0L;
	      long  lType  = 0L;
	      char *lValue = NULL;

	      recPrsGrsVarTriple(lNodeVar,&lClass,&lType,(void**)&lValue);

	      if ((lClass == tagVarClassPiece) &&
		  (lType == tagVarClassPieceTargetToken))
		
	       { 
		 *pTargetPtr = lValue;
		 break;
	       }

	      lNodeVar = lNodeVar->next;
	    }

	 }

	lTmp = lTmp->next;
      }
   }

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsVariant
 Purpose:    Parses records returned from a request for the "VARIANT"
             element set.
             On the client side, go through all tags, find out what tagSet
             they come from, then determine if the tag is a string or a numeric
	     tag.  If it's a numeric tag, see if we can recognize it (if it's
	     well-known).  if it's a string tag, put it in with additional
	     elements.
	     TagType is from one of: tagSetM (meta information)
	                             tagSetG (generic information)
		                     locally-defined
             If the tag comes from tagSetM or tagSetG, the tag
	     is most likely a numeric tag, and the value of the
	     tag is returned in the variable lTagValNum.
	     If the tag is a locally-defined tag (of TagType 3)
	     then the tag is a string tag, and the string should
	     be supplied in the variable lTagValStr.
 Called by:  recPrsGrs
 Parameters: pNode  -- input;
             pDiags -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsVariant(DATA_DIR* pNode, 
		 list*     pDiags,
		 va_list   va)
{ 
  list *pRecs = va_arg(va,list*);
  boolean lRetCode = true;

  /* The headline data structure. */
  appConnHeadline *lRec = NULL;

  /* Elements for the headline data structure lRec. */
  char *lHl    = NULL;
  char *lName  = NULL;
  date *lDate  = NULL;
  long  lRank  = 0L;
  long  lScore = 0L;
  any  *lRecId = NULL;
  char *lURx   = NULL;
  list  lElems = NULL;

  DATA_DIR *lNode = pNode;

  /* Loop over all the tagged elements of the record. */
  while (lNode != NULL)
   { 
     DATA_DIR *lTmp;
     long  lTagType = 0;
     long  lTagValType = 0;
     long  lTagValNum = 0;
     char *lTagValStr = NULL;
     long  lLen;

     lTmp = tagFindNext(lNode,tagUnivSeq);
     lTmp = lTmp->ptr.child;

     recPrsGrsTagType(lTmp,&lTagType);
     recPrsGrsTagVal(lTmp,&lTagValType,&lTagValNum,&lTagValStr);

     /* Dispatch for each tagged element */
     switch (lTagType)
      { 
      /* Meta Information Tag Set */
      case tagTagSetM:
       { 
	 switch (lTagValNum)
	  { 
	  case tagSetMRank:  recPrsGrsCntsInt(lTmp,&lRank); break;
	  case tagSetMScore: recPrsGrsCntsInt(lTmp,&lScore); break;
	  case tagSetMURx:   recPrsGrsCntsStr(lTmp,&lURx,&lLen); break;
	  case tagSetMRecId: 
	   {
	     char *lRecIdStr = NULL;
	     recPrsGrsCntsStr(lTmp,&lRecIdStr,&lLen);
	     lRecId = stringToAny(lRecIdStr);
	     safeFree(lRecIdStr);
	     break;
	   }
	  }
	 break;
       }

	/* Generic Information Tag Set */
      case tagTagSetG:	
       { 
	 switch (lTagValNum)
	  { 
	  case tagSetGHeadline: recPrsGrsCntsStr(lTmp,&lHl,&lLen); break;
	  case tagSetGName:     recPrsGrsCntsStr(lTmp,&lName,&lLen); break;
	  case tagSetGDate:     recPrsGrsCntsDate(lTmp,&lDate); break;
	  }
	 break;
       }

	/* Locally-defined Tag Set */
      case tagTagSetL:	
       {
	 appConnElement *lElem;
	 if (recPrsGrsMetaDataElem(lTmp,lTagValStr,&lElem))
	   lElems = collecting(lElems,lElem);
	 else lRetCode = false;
	 break;
       }

	/* Unknown Tag Set */
      default:
       {
	 lRetCode = false;
	 break;
       }
      }
      
     safeFree(lTagValStr);

     lNode = lNode->next;
   }

  /* Check for the mandatory elements. */
  if ((lHl != NULL) &&
      (lRecId != NULL))

    /* If the mandatory elements are found, create headline data structure. */
   { lRec = appConnMakeHeadline(lHl,lName,lDate,lRank,lScore,lRecId,lURx,
				lElems);
     *pRecs = collecting(*pRecs,lRec);

     safeFree(lHl);
     safeFree(lName);
     safeFree(lDate);
     safeFree(lRecId->bytes);
     freeAny(lRecId);
     safeFree(lURx);
   }
  else lRetCode = false;

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrsObjElem
 Purpose:    Parses records returned from a request for the "Object Element"
             element set.
 Called by:  recPrsGrs
 Parameters: pNode  -- input;
             pDiags -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsGrsObjElem(DATA_DIR* pNode,
		 list*     pDiags,
		 va_list   va)
{ 
  char **pTagValStr = va_arg(va,char**);
  char **pElem = va_arg(va,char**);
  long  *pLen = va_arg(va,long*);
  char **pTargetPtr = va_arg(va,char**);

  DATA_DIR *lNode = pNode;
  
  while (lNode != NULL)
    { 
      long  lTagType    = 0L;
      long  lTagValType = 0L;
      long  lTagValNum  = 0L;

      DATA_DIR *lTmp = tagFindNext(lNode,tagUnivSeq);
      lTmp = lTmp->ptr.child;

      recPrsGrsTagType(lTmp,&lTagType);
      recPrsGrsTagVal(lTmp,&lTagValType,&lTagValNum,pTagValStr);

      if ((lTagValType == tagGRSTagValStr))

	/* XXX This could come in any type. */
	recPrsGrsCntsStr(lTmp,pElem,pLen);

      /* Check for the target pointer. */
      recPrsGrsMetaTargetPtr(lTmp,pTargetPtr);

      lNode = lNode->next;
    }

  return(true);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsGrs
 Purpose:    Determines what kind of element set was requested in the record,
             and dispatches to the appropriate handler.
 Called by:  recPrsDbRec  
 Parameters: pDir   -- input;
             pDiags -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean 
recPrsGrs(DATA_DIR* pDir,
	  list*     pDiags,
	  va_list   va)
{
  long pElmtSet = va_arg(va,long);
  boolean lRetCode = false;

  switch (pElmtSet)
   {
   case VARIANT_ELEM_SET:
    { 
      lRetCode = recPrsGrsVariant(pDir,pDiags,va);
      break;
    }
   case OBJ_ELEM_ELEM_SET:
    {
      lRetCode = recPrsGrsObjElem(pDir,pDiags,va);
      break;
    }
   }  

  return(lRetCode);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsDbRec
 Purpose:    Parses a database record, determines what kind of records it is,
             and dispatches to the appropriate handler.  Currently, only GRS
             (generic record syntax) is supported.  Could easily be extended
             to handle other record syntaxes.
 Called by:  recPrsDbOrSurDiag
 Parameters: pDir   -- input;
             pDiags -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsDbRec(DATA_DIR* pDir,
	    list*     pDiags,
	    va_list   va)
{
  boolean lRetCode = false;
  DATA_DIR *lTmp = NULL;
  DATA_DIR *lOidDir = NULL;
  char *lOid = NULL;

  /* Check for an external definition and get the record syntax. */
  if (((lTmp = tagFindChild(pDir,tagUnivExt)) != NULL) &&
      ((lOidDir = tagFindChild(lTmp,tagUnivObjId)) != NULL))

   { 
     /* Find the Object Identifier for the Record Syntax. */
     lOid = dgetoid(lOidDir);

     /* Determine the Record Syntax. */
     if (!strncmp(lOid,OID_RECSYN_SUTRES,safeStrlen(OID_RECSYN_SUTRES)))
      {	
	/* Unstructured Text Record Syntax */ 
	if (((lTmp = tagFindChild(lTmp,tagExtEncSin)) != NULL) &&
	    ((lTmp = tagFindChild(lTmp,tagUnivGenStr)) != NULL))

	  return(false);
      }

     else if (!strncmp(lOid,OID_RECSYN_GENERIC,safeStrlen(OID_RECSYN_GENERIC)))
      { 
	/* Generic Record Syntax */
	long lTagType, lElemType;
	if (((lTmp = tagFindChild(lTmp,tagExtEncSin)) != NULL) &&
	    ((lTmp = lTmp->ptr.child) != NULL))

	  lRetCode = recPrsGrs(lTmp,pDiags,va);
      }

     
     else if (!strncmp(lOid,OID_RECSYN_USMARC,safeStrlen(OID_RECSYN_USMARC)))
      { 
	/* USMARC Record Syntax */
	if ((lTmp = tagFindChild(lTmp,tagExtEncOctAlign)) != NULL)

	  return(false);
      }
   }

  return(lRetCode);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsDiag
 Purpose:    Parses the diagnostic in pDir, collects it in the pDiags list.
             The diagnostic may be surrogate or non-surrogate.
 Called by:  recPrsRecs, recPrsDbOrSurDiag
 Parameters: pDir   -- input;
             pDiags -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsDiag(DATA_DIR* pDir,
	   list*     pDiags)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp = NULL;
  char *lOid = NULL;
  long lCond = 0L;
  char lAddnInfo[MAX_STR_LEN];
  
  if ((lTmp = tagFindChild(pDir,tagUnivObjId)) != NULL)
    lOid = dgetoid(lTmp);
  else lRetCode = false;

  if ((lTmp = tagFindNext(lTmp,tagUnivInt)) != NULL)
    lCond = dgetnum(lTmp);
  else lRetCode = false;

  if (((lTmp = tagFindChild(pDir,tagUnivVisStr)) != NULL) ||
      ((lTmp = tagFindChild(pDir,tagUnivGenStr)) != NULL))
    safeStrncpy(lAddnInfo,(char*)lTmp->ptr.data,
		MIN(lTmp->count+1L,MAX_STR_LEN));
  else lRetCode = false;

  /* Remember the diagnostic. */
  if (lRetCode != false)
    { 
      appConnDiag *lDiag;

      /* Check for the Bib-1 Diagnostic Set. */
      if (!(strncmp(lOid,OID_DIAG_BIB1,safeStrlen(OID_DIAG_BIB1))))
       {
	 long  lAppConnDiagCode = diagCondAppConnDiag(lCond);
	 char* lInfoFormat = "diagnostic received from Z39.50 Diagnostic-1 Set with condition: %s, additional info = %s";
	 char* lDiagMeaning = diagCondMeaning(lCond);

	 lDiag = appConnDiagMake(lAppConnDiagCode,lInfoFormat,
				 lDiagMeaning,lAddnInfo);
       }

      else
	/* If it is not from the Diag1 Set. */
	lDiag = appConnDiagMake(appConnDiagProtocolError,"diagnostic received with unrecognized diagnostic object identifier %s, condition %ld, additional info = %s.",lOid,lCond,lAddnInfo);
	 
      *pDiags = collecting(*pDiags,lDiag);
    }

  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsDbOrSurDiag
 Purpose:    Loops over and parses a sequence of records and possibly 
             surrogate diagnostics.  Captures surrogate or non-surrogate
             diagnostics in the pDiags list.
 Called by:  recPrsRecs
 Parameters: pDir   -- input;
             pDiags -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
recPrsDbOrSurDiag(DATA_DIR* pDir,
		  list*     pDiags,
		  va_list   va)
{
  boolean lRetCode = true;
  DATA_DIR *lRec = NULL;
  DATA_DIR *lTmp = NULL;

  /* Look for the database Record or the surrogate Diagnostic. */
  if ((lRec = tagFindChild(pDir,tagRecNamePlusRecDbRecOrSurDiag)) != NULL)

   { 
     if ((lTmp = tagFindChild(lRec,tagRecNamePlusRecDbRec)) != NULL)

      {
	/* Found a record. */
	lRetCode = recPrsDbRec(lTmp,pDiags,va);
      }

     else if ((lTmp = tagFindChild(lRec,tagRecNonSurDiag)) != NULL)

      {
	/* Found a surrogate diagnostic. */
	lRetCode = recPrsDiag(lTmp,pDiags);
      }

   }

  return(lRetCode);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   recPrsRecs
 Purpose:    Parse the records in the pDir directory tree, and create output
             data structures appropriate to the element set requested.  Parses
             either a non-surrogate diagnostic or a sequence of records and
             possibly surrogate diagnostics. Captures surrogate or non-
             surrogate diagnostics in the pDiags list.
 Called by:  Z3950V2SearchRespReadVariantsAPDU
             Z3950V2SearchRespReadObjElemAPDU
             Z3950V2PresentRespReadVariantsAPDU
             Z3950V2PresentRespReadObjElemAPDU
 Parameters: pDir   -- input;
             pDiags -- output;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
recPrsRecs(DATA_DIR* pDir,
	   list*     pDiags,...)
{
  boolean lRetCode = true;
  DATA_DIR *lTmp;

  va_list ap;
  va_start(ap,pDir);

  /* Get the database or surrogate diagnostic records. */
  if ((lTmp = tagFindChild(pDir,tagRecDbOrSurDiag)) != NULL)

   {				
     /* Get the first record. */
     if ((lTmp = tagFindChild(lTmp,tagUnivSeq)) != NULL)

      {
	/* Loop over the each record and parse. */
	while(lTmp != NULL)
	 { 

	   /* Parse the record. */
	   if (!(lRetCode = recPrsDbOrSurDiag(lTmp,pDiags,ap)))
	     break;
	
	   /* Go to the next record. */
	   lTmp = tagFindNext(lTmp->next,tagUnivSeq);

	 }
      }

     /* Else no records were sent. */
   }

  /* Check for a non-surrogate database record. */
  else if ((lTmp = tagFindChild(pDir,tagRecNonSurDiag)) != NULL)

    /* Parse the non-surrogate diagnostic. */
    lRetCode = recPrsDiag(lTmp,pDiags);

  va_end(ap); 

  if (!lRetCode)
    /* Unable to read the records, create a diagnostic. */
   { 
     appConnDiag *lDiag = 
       appConnDiagMake(appConnDiagBadPacket,
		       "unable to parse database records");
     *pDiags = collecting(*pDiags,lDiag);
   }

  return lRetCode;
}
