/*****************************************************************************
* 	        (c) Copyright 1994 Wide Area Information Servers, Inc        *
*	   	                 All rights reserved.   	             *
*									     *
*  This notice is intended as a precaution against inadvertent publication   *
*  and does not constitute an admission or acknowledgement that publication  *
*  has occurred or constitute a waiver of confidentiality.		     *
*									     *
*  Wide Area Information Server software is the proprietary and              *
*  confidential property of Wide Area Information Servers, Inc.              *
*****************************************************************************/

/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Module:   rpnBld.c

 Purpose:  This module provides functions for building reverse-polish notation
           (RPN) Type-1 Z39.50 queries using attributes from the Bib-1 
           Attribute Set.  The top-level function is called rpnBld().
---------------------------------------------------------------------------*/

#include <string.h>
#include "cutil.h"
#include "appConnMgr.h"

#include "berutil.h"
#include "tag.h"
#include "attr.h"
#include "rpnBld.h"



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   rpnBldAttrElmt
 Purpose:    Build an attribute element of a query.  The result contains
             a sequence of Attribute, Value pairs.  The resulting structure
             is stored in the parameter pNode.
 Called by:  rpnBldInternal
 Parameters: pNode     -- input and output;
             pAttrType -- input;
             pAttrVal  -- input;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
rpnBldAttrElmt(DATA_DIR* pNode,
	       long      pAttrType,
	       long      pAttrVal)
{ 
  boolean lRetCode = true;
  DATA_DIR *lNode = NULL;

  /* Next term in sequence */
  lNode = daddtag(pNode,tagUnivSeq,ASN1_UNIVERSAL);

  /* Term attribute */
  daddnum(lNode,tagRPNAttrType,ASN1_CONTEXT,
	  (CHAR*)&pAttrType,sizeof(long));

  /* Term value */
  daddnum(lNode,tagRPNAttrVal,ASN1_CONTEXT,
	  (CHAR*)&pAttrVal,sizeof(long));

  return(true);
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   rpnBldInternal
 Purpose:    Recursively builds a Type-1 Query from pQry.  The resulting 
             structure is stored in the parameter pNode.
 Called by:  rpnBld, rpnBldInternal
 Parameters: pNode -- input and output;
             pQry  -- input;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
static boolean
rpnBldInternal(DATA_DIR*   pNode,
	       appConnQry* pQry)
{ 
  boolean lRetCode = true;
  DATA_DIR *lNode = NULL, *lAttrPlusTerm;
  
  /* If the query only contains a single term. */
  if (pQry->term != NULL)
   { any *lTerm = pQry->term;

     /* Grab the attributes from the attribute table. */
     tAttrList lAttrList = gAttrTable[pQry->type];

     /* Create an rpn operand node. */
     lNode = daddtag(pNode,tagRPNOperand,ASN1_CONTEXT);

     /* Create an attribute list plus term node. */
     lAttrPlusTerm = daddtag(lNode,tagRPNAttrPlusTerm,ASN1_CONTEXT);

     /* Create the attribute list node. */
     lNode = daddtag(lAttrPlusTerm,tagRPNAttrList,ASN1_CONTEXT);

     /* Create the attribute elements with attribute types
	of Use, Relation, and Structure. */
     rpnBldAttrElmt(lNode,attrTypeUse,   lAttrList.sAttrUse);
     rpnBldAttrElmt(lNode,attrTypeRelat, lAttrList.sAttrRelat);
     rpnBldAttrElmt(lNode,attrTypeStruct,lAttrList.sAttrStruct);   

     /* Create the term. */
     daddchar(lAttrPlusTerm,tagRPNTerm,ASN1_CONTEXT,
	      (CHAR*)lTerm->bytes,lTerm->size);
   }   

  /* If the query is made up of two subqueries and an operator. */
  else 
   { lNode = daddtag(pNode,tagRPNSeq,ASN1_CONTEXT);

     /* Recurse over operands in query. */
     rpnBldInternal(lNode,pQry->arg1);
     rpnBldInternal(lNode,pQry->arg2);

     /* Create the rpn operator node. */
     lNode = daddtag(lNode,tagRPNOperator,ASN1_CONTEXT);
     daddtag(lNode,pQry->operator,ASN1_CONTEXT);
  }
  return lRetCode;
}



/*---------------------------------------------------------------------------
  ---------------------------------------------------------------------------
 Function:   rpnBldBer
 Purpose:    Top level build function for creating a Type-1 RPN query
             from a pQry.  It uses the Bib-1 Atrribute Set.  The resulting 
             structure is stored in the parameter pNode.
 Called by:  Z3950V2SearchReqWriteAPDU
 Parameters: pDir -- input and output;
             pQry -- input;
 Returns:    true(1) or false(0) depending on whether the function succeeds
             or fails.
---------------------------------------------------------------------------*/
boolean
rpnBld(DATA_DIR*   pDir,
       appConnQry* pQry) 
{ 
  boolean lRetCode = true;
  DATA_DIR *lNode = NULL;
  
  /* Create the top-level query node. */   
  lNode = daddtag(pDir,tagSrchQry,ASN1_CONTEXT);

  /* Create the type-1 rpn query node. */
  lNode = daddtag(lNode,tagRPNQuery,ASN1_CONTEXT);

  /* Use the ANSI Z39.50 OID for Bib-1. */
  daddoid(lNode,tagUnivObjId,ASN1_UNIVERSAL,OID_ATTRSET_BIB1);

  /* Recursively build the rpn structure. */ 
  rpnBldInternal(lNode,pQry);

  return lRetCode;
}
